##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Demonstration and test flowsheet for a dynamic flowsheet.

"""
from __future__ import division
from __future__ import print_function
import matplotlib.pyplot as plt

# Import Pyomo libraries
from pyomo.environ import ConcreteModel, Constraint, SolverFactory, \
                            TransformationFactory, Var

# Import IDAES core
from idaes.core import FlowsheetBlock, Stream

# Import Unit Model Modules
import tutorial_1_properties as properties_rxn

# Import Unit Model Modules
from idaes.models import CSTR, Mixer


def main():
    """
    Make the flowsheet object, fix some variables, and solve the problem
    """
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=True,
                          time_set=[0, 1, 500000])

    # Add property packages to flowsheet library
    m.fs.properties_rxn = properties_rxn.PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.properties_rxn

    # Create unit models
    m.fs.Mix = Mixer(dynamic=False)
    m.fs.Tank1 = CSTR()
    m.fs.Tank2 = CSTR()

    # Add constraints to Tank1
    m.fs.Tank1.height = Var(m.fs.Tank1.time,
                            initialize=1.0,
                            doc="Depth of fluid in tank [m]")
    m.fs.Tank1.area = Var(initialize=1.0,
                          doc="Cross-sectional area of tank [m^2]")
    m.fs.Tank1.volume_flow = Var(m.fs.Tank1.time,
                                 initialize=4.2e5,
                                 doc="Volumetric flow leaving tank")
    m.fs.Tank1.flow_coeff = Var(m.fs.Tank1.time,
                                initialize=5e-5,
                                doc="Tank outlet flow coefficient")

    def geometry(b, t):
        return b.volume[t] == b.area*b.height[t]
    m.fs.Tank1.geometry = Constraint(m.fs.Tank1.time, rule=geometry)

    def volume_flow_calculation(b, t):
        return b.volume_flow[t] == (
                    b.holdup.properties_out[t].flow_mol /
                    b.holdup.properties_out[t].dens_mol_phase['Liq'])
    m.fs.Tank1.volume_flow_calculation = Constraint(
            m.fs.Tank1.time,
            rule=volume_flow_calculation)

    def outlet_flowrate(b, t):
        return b.volume_flow[t] == b.flow_coeff[t]*b.height[t]
    m.fs.Tank1.outlet_flowrate = Constraint(m.fs.Tank1.time,
                                            rule=outlet_flowrate)

    # Add constraints to Tank2
    m.fs.Tank2.height = Var(m.fs.Tank2.time,
                            initialize=1.0,
                            doc="Depth of fluid in tank [m]")
    m.fs.Tank2.area = Var(initialize=1.0,
                          doc="Cross-sectional area of tank [m^2]")
    m.fs.Tank2.volume_flow = Var(m.fs.Tank2.time,
                                 initialize=4.2e5,
                                 doc="Volumetric flow leaving tank")
    m.fs.Tank2.flow_coeff = Var(m.fs.Tank2.time,
                                initialize=5e-5,
                                doc="Tank outlet flow coefficient")

    m.fs.Tank2.geometry = Constraint(m.fs.Tank2.time, rule=geometry)
    m.fs.Tank2.volume_flow_calculation = Constraint(
            m.fs.Tank2.time,
            rule=volume_flow_calculation)
    m.fs.Tank2.outlet_flowrate = Constraint(m.fs.Tank2.time,
                                            rule=outlet_flowrate)

    # Apply DAE Transformation
    discretizer = TransformationFactory('dae.finite_difference')
    discretizer.apply_to(m.fs,
                         nfe=200,
                         wrt=m.fs.time,
                         scheme='BACKWARD')

    # Post transform build
    m.fs.post_transform_build()

    # Make Streams
    m.fs.stream_1 = Stream(source=m.fs.Mix.outlet,
                           destination=m.fs.Tank1.inlet)

    m.fs.stream_2 = Stream(source=m.fs.Tank1.outlet,
                           destination=m.fs.Tank2.inlet)

    # Set inlet and operating conditions, and some initial conditions.
    m.fs.Mix.inlet[:, "1"].vars["flow_mol_comp"]["a"].fix(10.0)
    m.fs.Mix.inlet[:, "1"].vars["flow_mol_comp"]["b"].fix(0.0)
    m.fs.Mix.inlet[:, "1"].vars["flow_mol_comp"]["c"].fix(1.0)
    m.fs.Mix.inlet[:, "1"].vars["flow_mol_comp"]["d"].fix(0.0)
    m.fs.Mix.inlet[:, "1"].vars["flow_mol_comp"]["e"].fix(0.0)
    m.fs.Mix.inlet[:, "1"].vars["flow_mol_comp"]["f"].fix(0.0)
    m.fs.Mix.inlet[:, "1"].vars["temperature"].fix(303.15)
    m.fs.Mix.inlet[:, "1"].vars["pressure"].fix(101325.0)

    m.fs.Mix.inlet[:, "2"].vars["flow_mol_comp"]["a"].fix(0.0)
    m.fs.Mix.inlet[:, "2"].vars["flow_mol_comp"]["b"].fix(20.0)
    m.fs.Mix.inlet[:, "2"].vars["flow_mol_comp"]["c"].fix(0.0)
    m.fs.Mix.inlet[:, "2"].vars["flow_mol_comp"]["d"].fix(0.0)
    m.fs.Mix.inlet[:, "2"].vars["flow_mol_comp"]["e"].fix(0.0)
    m.fs.Mix.inlet[:, "2"].vars["flow_mol_comp"]["f"].fix(0.0)
    m.fs.Mix.inlet[:, "2"].vars["temperature"].fix(303.15)
    m.fs.Mix.inlet[:, "2"].vars["pressure"].fix(101325.0)

    m.fs.Tank1.area.fix(0.5)
    m.fs.Tank1.flow_coeff.fix(5e-6)
    m.fs.Tank1.heat.fix(0.0)

    m.fs.Tank2.area.fix(0.5)
    m.fs.Tank2.flow_coeff.fix(5e-6)
    m.fs.Tank2.heat.fix(0.0)

    # Set Initial Conditions
    m.fs.fix_initial_conditions('steady-state')

    # Initialize Units
    m.fs.Mix.initialize()
    m.fs.Tank1.initialize(state_args={
            "flow_mol_comp": {
                "a": m.fs.Mix.outlet[0].vars["flow_mol_comp"]["a"].value,
                "b": m.fs.Mix.outlet[0].vars["flow_mol_comp"]["b"].value,
                "c": m.fs.Mix.outlet[0].vars["flow_mol_comp"]["c"].value,
                "d": m.fs.Mix.outlet[0].vars["flow_mol_comp"]["d"].value,
                "e": m.fs.Mix.outlet[0].vars["flow_mol_comp"]["e"].value,
                "f": m.fs.Mix.outlet[0].vars["flow_mol_comp"]["f"].value},
            "pressure": m.fs.Tank1.outlet[0].vars["pressure"].value,
            "temperature": m.fs.Tank1.outlet[0].vars["temperature"].value})
    m.fs.Tank2.initialize(state_args={
            "flow_mol_comp": {
                "a": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["a"].value,
                "b": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["b"].value,
                "c": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["c"].value,
                "d": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["d"].value,
                "e": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["e"].value,
                "f": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["f"].value},
            "pressure": m.fs.Tank1.outlet[0].vars["pressure"].value,
            "temperature": m.fs.Tank1.outlet[0].vars["temperature"].value})

    # Create a solver
    solver = SolverFactory('ipopt')
    results = solver.solve(m, tee=True)

    # Create a disturbance
    for t in m.fs.time:
        if t >= 1.0:
            m.fs.Mix.inlet[t, "2"].vars["flow_mol_comp"]["b"].fix(10.0)

    results = solver.solve(m, tee=True)

    # Print results
    print(results)
    print()
    print("Results")
    print()

    # For testing purposes
    return(m, results)


def plot_results(m):
    time = []
    a = []
    b = []
    c = []
    d = []
    e = []
    f = []
    for t in m.fs.time:
        time.append(t)
        a.append(m.fs.Tank2.outlet[t].vars["flow_mol_comp"]["a"].value)
        b.append(m.fs.Tank2.outlet[t].vars["flow_mol_comp"]["b"].value)
        c.append(m.fs.Tank2.outlet[t].vars["flow_mol_comp"]["c"].value)
        d.append(m.fs.Tank2.outlet[t].vars["flow_mol_comp"]["d"].value)
        e.append(m.fs.Tank2.outlet[t].vars["flow_mol_comp"]["e"].value)
        f.append(m.fs.Tank2.outlet[t].vars["flow_mol_comp"]["f"].value)

    plt.figure(1)
    plt.plot(time, a, label='a')
    plt.plot(time, b, label='b')
    plt.plot(time, c, label='c')
    plt.plot(time, d, label='d')
    plt.plot(time, e, label='e')
    plt.plot(time, f, label='f')
    plt.legend()
    plt.grid()
    plt.xlabel("Time [s]")
    plt.ylabel("Molar Flowrate [mol/s]")
    plt.show(block=True)


if __name__ == "__main__":
    m, results = main()
    plot_results(m)
