#!/usr/bin/env python
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Create and populate a DMF workspace with a prototypical
set of resources.

    [results]                      [results]
     | experiment                     | experiment
     v                               v
    [experiment 2] -- version --> [experiment 1]
      | experiment                    | experiment
      v                               v
    [flowsheet 1.2.0] -- version --> [flowsheet 1.1.0]
      (params, code)   _____________/     (params, code)
      |               /
      v contains  <-+
    [properties]

"""
# stdlib
import argparse
import logging
import os
from random import randint
import shutil
import sys
import tempfile
# third-party
import pendulum
from six.moves import input
# package
from idaes.dmf.dmf import DMF
from idaes.dmf import resource, experiment
from idaes.dmf.resource import R_CONTAINS, Version
from idaes.dmf.util import get_logger

_log = get_logger('generate_workspace')


def main():
    # argument parsing
    prs = argparse.ArgumentParser()
    prs.add_argument('workspace')
    prs.add_argument('-n', '--name', dest='name', default='test',
                     help='Workspace name (default=test)')
    prs.add_argument('-d', '--desc', dest='desc', default='Test workspace',
                     help='Workspace description')
    prs.add_argument('-v', '--verbose', dest='vb', action='count',
                     default=0)
    args = prs.parse_args()
    if args.vb > 1:
        _log.setLevel(logging.DEBUG)
    elif args.vb > 0:
        _log.setLevel(logging.INFO)
    else:
        _log.setLevel(logging.WARN)
    ws_path = args.workspace
    # remove existing workspace if any
    if os.path.exists(ws_path):
        q = input('Remove existing workspace at {} (Y/n)? '.format(ws_path))
        if q.lower().startswith('n'):
            _log.warn('Abort')
            return 1
        else:
            shutil.rmtree(ws_path)
            _log.info('Removed old workspace')
    _log.info('Creating new workspace')
    # create workspace
    num_runs = 2
    runs = list(range(1, num_runs + 1))
    # create resources, except experiments
    prop = resource_properties()
    ofiles, results = [], []
    for rnum in runs:
        f = tempfile.NamedTemporaryFile()
        results.append(resource_results(rnum, f))
        ofiles.append(f)
    flowsheets = list(map(resource_flowsheets, runs))
    # initialize DMF
    _log.debug('Initializing DMF')
    dmf = DMF(path=ws_path, create=True, name=args.name, desc=args.desc,
              add_defaults=True)
    # create experiments
    _log.debug('Creating experiments')
    for i in runs:
        add_new_rsrc = i == runs[0]
        xp = experiment.Experiment(dmf, name='experiment {}'.format(i))
        xp.add(prop)
        xp.add(results[i - 1])
        xp.add(flowsheets[i - 1])
        _log.debug('Experiment {} created'.format(i))
    # add relations
    _log.debug('Adding relations')
    for i in runs:
        flow = flowsheets[i - 1]
        flow.create_relation(flow, R_CONTAINS, prop)
        dmf.update(flow)
    _log.info('Done')
    print('Created new workspace at {}'.format(ws_path))
    return 0


def resource_results(num, datafile):
    r = resource.Resource(type=resource.ResourceTypes.data)
    r.created = _get_time()
    r.desc = 'Result set'
    r.name = 'results'
    r.version = Version(revision=1)
    r.creator = resource.Contact(name='Some Person',
                                 email='person@local.org')
    r.datafiles = [resource.FilePath(path=datafile.name,
                                     desc='Output data')]
    return r


def resource_properties():
    r = resource.Resource(type=resource.ResourceTypes.property_data)
    r.created = _get_time()
    r.desc = 'Result set'
    r.name = 'results'
    return r


def resource_experiments(num):
    r = resource.Resource(type=resource.ResourceTypes.experiment)
    r.created = _get_time()
    r.version = Version(revision=num)
    r.desc = 'An experimental run'
    r.name = 'run'
    return r


def resource_flowsheets(num):
    r = resource.Resource(type=resource.ResourceTypes.fs)
    r.version = Version(revision=(1, num, 0))
    r.created = _get_time()
    r.desc = 'The main flowsheet'
    r.name = 'main flowsheet'
    return r


def _get_time():
    return pendulum.datetime.now()

if __name__ == '__main__':
    sys.exit(main())