##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Simple MEA Column unit model. This model uses a finite volume 
approximation, offest between the liquid (forward differnce) and vapor 
(bacward differnce) can be confusing.

Art Time:

  Liquid (Backward dif)->--+   +----> Vapor (Forward dif)
                           |   |
                          +-----+ --- 0,  z=0
                          |\   /|
                          | \ / |
                          |  X  |
                          | / \ |
                          |/   \|
                          +-----+ --- L, z=1.0
                           |   |
               Liquid <----+   +---<- Vapor
    
    z: dimensionless distance in axial direction
    L: column length
    L*z: distance in axial direction
    
    The column height can be changed without changing the discretiation
    by changing L.
    
Discretization Scheme:

    Points               1   2   3 ... n-1    n
    Volumes Liquid  1 BC | 2 | 3 | ...  |  n  |       flow -->
                            /  |
             Driving force /   |
                          /    | Heat/mass transfer
                         /     |
    Volume Vapor         | 1 | 2 | ...  | n-1 | n BC  <-- flow
    Points               1   2   3 ... n-1    n
    
    The volume elements are indexed by the point that represents 
    the balance on the volume. The conditions in the volume are
    assumed to be the same as at the exit of the volume.
    
    In the liquid phase volume 1 doesn't exist, and point 1 is a 
    boundary, In the vapor phase volume n does not exist, and 
    point n is a boundary.  This means that the liqid and vapor 
    phases line up lide this:
    
       Liquid Points:  BC1  2   3 ... n-1    n
       Liquid Volumes:  | 2 | 3 | ...  |  n  | --> Flow
       Vapor Volumes    | 1 | 2 | ...  | n-1 | <-- Flow
       Liquid Volumes:  1   2   3     n-1   BCn
       
    The mass/energy transfer is between volume n in the liquid and n-1
    in the vapor.  The driving force is between the volume outlets, 
    which is between n in the liquid and n-1 in the vapor.
"""
from __future__ import division
from __future__ import print_function
__author__ = "John Eslick"
__version__ = "1.0.0"

from idaes_models.core.unit_model import *
from pyomo.environ import *
from pyomo.dae import *
from pyomo.dae.plugins.finitedifference import \
    Finite_Difference_Transformation
from transport import Transport
import math
import time
import six
import idaes_models.core.model_util as model_util
import pandas
import pyutilib.services

__all__ = ['Absorber']

class Absorber(UnitModel):
    __geometry__ =\
    """
  Liquid (Backwar dif)->--+   +----> Vapor (Forward dif)
                          |   |
                         +-----+ --- 0,  z=0
                         |\   /|
                         | \ / |
                         |  X  |
                         | / \ |
                         |/   \|
                         +-----+ --- L, z=1.0
                          |   |
              Liquid <----+   +---<- Vapor
"""

    def __init__(self, nfe=30, main_model=None):
        """
        Args:
        nfe = number of finite elements for column discretization
        """
        UnitModel.__init__(self, index_set=None)
        self.nfe = nfe
        # Some mole fraction values for testing/intialization
        self.x_init = {}
        self.y_init = {}
        self.T_init = 320
        self.P_init = 101325
        self.x_init['CO2'] = 0.0110
        self.x_init['H2O'] = 0.8780
        self.x_init['MEA'] = 0.1110
        self.y_init['CO2'] = 0.1265
        self.y_init['H2O'] = 0.1233
        self.y_init['MEA'] = 0.0000
        self.y_init['N2'] = 0.7502
        self.transport = None # Transport property pack
        self._reg_initialize(
            'default', func=self._default_init, 
            doc='Default absorber initialization')
        self.main_model = main_model
        
    def make_profile(self):
        """
        """
        blk = self.pyomo_block
        zset = blk.z
        z = [z*blk.L.value for z in zset]
        p_V_vap = self.main_model.p_V_vap
        self.profile_1 = pandas.DataFrame(
            data={
                'z':z,
                'T_liq':[blk.T_l[z].value for z in zset],
                'T_vap':[blk.T_v[z].value for z in zset],
                'y_vap_CO2':[blk.y_v[z,'CO2'].value for z in zset],
                'y_liq_CO2':[blk.y_l[z,'CO2'].value for z in zset]})
        
    def __call__(self, blk, index=0):
        self.pyomo_block = blk
        self._make_params(blk)
        self._make_vars(blk)
        self._make_derivatives(blk)
        self._fd_transform(blk)
        self._adjust_bounds(blk)
        self._make_transport(blk)
        self._chem_eq(blk)
        self._mass_balance(blk)
        self._energy_balance(blk)
        self._make_ports(blk)
        return blk
        
    def _make_params(self, blk):
        """
        Create the paramters and sets for the Pyomo block
        """
        # Initialization parameters used to turn heat and mass transfer
        # up and down
        blk.init_mt_scale = Param(initialize=1, mutable=True)
        blk.init_ht_scale = Param(initialize=1, mutable=True)
        
        # Component sets
        #
        # All the components in the system
        blk.comp = Set(initialize=['CO2', 'H2O', 'N2', 'MEA'])
        # Liquid phase components (assuming no N2 in solvent)
        blk.lcomp = Set(initialize=['CO2', 'H2O', 'MEA'])
        # Henry's law components
        blk.hcomp = Set(initialize=['CO2'])
        # Raoult's law components
        blk.vpcomp = Set(initialize=['H2O', 'MEA'])
        # Chemical components in solvent
        blk.ccomp = Set(initialize=['CO2', 'OH-', 'H3O+', 'HCO3-', 
                                    'CO3--', 'RNH2', 'RNHCOO-',
                                    'RNH3+'])
        blk.rxns = Set(initialize=[1,2,3,8,9])
        blk.z = ContinuousSet(bounds=(0,1)) #axial dimentionless
        blk.z_fwd = ContinuousSet(bounds=(0,1)) #axial dimentionless
        
    def _make_vars(self, blk):
        """
        Create the model variables.  At this point I'm using some 
        indexes in by dinite dfference approxiamtion so I just do the
        finite differnce transformation right away, instead of waiting
        until the end, like I probably should.
        """
        blk.L = Var(
            domain=Reals, bounds=(0,20), initialize=10,
            doc = "Column bed length (m)")
        blk.D = Var(
            domain=Reals, bounds=(0,10), initialize=0.6403,
            doc = "Column diameter (m)")
        blk.A = Var(
            domain=Reals, bounds=(0,100), initialize=0.5,
            doc = "Column corss-sectional area (m^2)")
        blk.F_vap_in = Var(
            domain=Reals, bounds=(0,100), initialize=0.5,
            doc = "Vapor flow in (mol/s)")
        blk.F_vap_out = Var(
            domain=Reals, bounds=(0,100), initialize=0.5,
            doc = "Vapor flow out (mol/s)")
        blk.F_liq_in = Var(
            domain=Reals, bounds=(0,100), initialize=0.5,
            doc = "Liquid flow in (mol/s)")
        blk.F_liq_out = Var(
            domain=Reals, bounds=(0,100), initialize=0.5,
            doc = "Liquid flow out (mol/s)")
        blk.y_vap1 = Var(
            blk.comp,
            domain=Reals, bounds=(0,1), initialize=0.5,
            doc = "Vapor composition in (mol/s)")
        blk.y_vap2 = Var(
            blk.comp,
            domain=Reals, bounds=(0,1), initialize=0.5,
            doc = "Vapor composition out (mol/s)")
        blk.y_liq1 = Var(
            blk.lcomp,
            domain=Reals, bounds=(0,1), initialize=0.5,
            doc = "Liquid composition in (mol/s)")
        blk.y_liq2 = Var(
            blk.lcomp,
            domain=Reals, bounds=(0,1), initialize=0.5,
            doc = "Liquid composition out (mol/s)")
        blk.uv = Var(
            blk.z_fwd, domain=Reals,
            bounds=(1e-1,3), initialize=0.8,
            doc="Apparent vapor velocity (m/s)")
        blk.ul = Var(
            blk.z, domain=Reals,
            bounds=(1e-4,0.01), initialize=0.002,
            doc="Apparent liquid velocity (m/s)")
        blk.Cu_v = Var(
            blk.z_fwd, blk.comp, domain=Reals,
            bounds=(0,80), initialize=1,
            doc="Vapor (component concentration)*(velocity) (mol/m2-s)")
        blk.Cu_l = Var(
            blk.z, blk.lcomp, domain=Reals,
            bounds=(1e-3,200), initialize=10,
            doc="Liquid (component concentration)*(velocity) (mol/m2-s)")
        blk.T_l = Var(
            blk.z, domain=Reals, bounds=(270,400),
            initialize=300, doc="Liquid temperature (K)")
        blk.T_v = Var(
            blk.z_fwd, domain = Reals, bounds=(270,400), 
            initialize=300, doc="Vapor temperature (K)")
        blk.P = Var(
            blk.z, domain = Reals, bounds=(1000, 1e8), 
            initialize=101325, doc="Pressure (Pa)")
        blk.y_v = Var(
            blk.z, blk.comp, domain = Reals, bounds=(0,1), 
            initialize=0.1, doc="Vapor mole fractions (K)")
        blk.y_l = Var(
            blk.z, blk.comp, domain = Reals, bounds=(0,1), 
            initialize=0.1, doc="Liquid mole fractions (K)")
        blk.logC = Var(
            blk.z, blk.ccomp, domain=Reals, bounds=(-25,20), 
            initialize=0,
            doc="log_e concentration chemical species in solvent")
        blk.C = Var(
            blk.z, blk.ccomp, domain=Reals, bounds=(0, 1e4), 
            initialize=1,
            doc="concentration chemical species in solvent mol/m3")
        blk.y_star = Var(
            blk.z, blk.lcomp, domain=Reals,
            bounds=(0.0,0.5), initialize=0.1,
            doc="Interface component concentations")
        blk.dy = Var(
            blk.z, blk.lcomp, domain=Reals,
            bounds=(-0.4,0.4), initialize=0,
            doc="Mole fraction differnce between vapor interface and "\
                "vapor bulk")
        blk.N_v = Var(
            blk.z, blk.lcomp, domain=Reals,
            bounds=(-1e3, 1e3), initialize=0,
            doc="Moles i transfered into vapor (mol/m3-s")
        blk.N_vtot = Var(
            blk.z, domain=Reals,
            bounds=(-1e3, 1e3), initialize=0,
            doc="Moles total transfered into vapor (mol/m3-s") 
        blk.Q_v = Var(
            blk.z, domain=Reals, bounds=(-2e6, 2e6), initialize=0,
            doc="Heat into liquid phase Watt/m3")
        blk.deltaT = Var(
            blk.z, domain=Reals, bounds=(-300, 300), initialize=0,
            doc="Temperature difference between vapor and liquid (K)")
        blk.h_abs_vap = Var(
            blk.z, domain=Reals, bounds=(-1e8, 1e8), initialize=0,
            doc="Heat of absorption and vaporization (Watt/m3)")

    def _make_ports(self, blk):
        """
        Make ports for inlet and outlet streams.
        """
        blk.liq_in = self.one_phase_fluid_port(
            blk.F_liq_in, blk.T_l[0], blk.P[0], blk.y_liq1)
        blk.liq_out = self.one_phase_fluid_port(
            blk.F_liq_out, blk.T_l[1], blk.P[1], blk.y_liq2)
        blk.vap_in = self.one_phase_fluid_port(
            blk.F_vap_in, blk.T_v[1], blk.P[1], blk.y_vap1)
        blk.vap_out = self.one_phase_fluid_port(
            blk.F_vap_out, blk.T_v[0], blk.P[0], blk.y_vap2)
            
        self.liq_in = blk.liq_in
        self.liq_out = blk.liq_out
        self.vap_in = blk.vap_in
        self.vap_out = blk.vap_out

    def _make_derivatives(self, blk):
        """
        Add the derivative variables to the Pyomo block
        """
        blk.dCu_vdz = DerivativeVar(
            blk.Cu_v, wrt=blk.z_fwd, bounds=(-1e3, 1e3), initialize=0)
        blk.dCu_ldz = DerivativeVar(
            blk.Cu_l, wrt=blk.z, bounds=(-1e4, 1e4), initialize=0)
        blk.dT_vdz = DerivativeVar(
            blk.T_v, wrt=blk.z_fwd, bounds=(-1e4, 1e4), initialize=0)
        blk.dT_ldz = DerivativeVar(
            blk.T_l, wrt=blk.z, bounds=(-1e4, 1e4), initialize=0)
        
    def _fd_transform(self, blk):
        """
        Do the finite difference transformation.
        """
        discretize = Finite_Difference_Transformation()
        discretize.apply_to(
            blk, nfe=self.nfe, wrt=blk.z, scheme='BACKWARD')
        discretize.apply_to(
            blk, nfe=self.nfe, wrt=blk.z_fwd, scheme='FORWARD')

    def _adjust_bounds(self, blk):
        """
        Make some adjustments to variable bounds.  Theses are just
        subsets of indexed variables, otherwise the bounds could have
        been set originally.
        """
        pass
        
    def _make_transport(self, blk):
        """
        Make the transoport model
        """
        self.transport = Transport(
            comp=blk.comp, lcomp=blk.lcomp, index_set=blk.z, 
            P=blk.P, T_L=blk.T_l, y_L = blk.y_l, T_V=blk.T_v, y_V=blk.y_v,
            u_liq=blk.ul, u_vap=blk.uv, logC=blk.logC, ystar=blk.y_star,
            packing='mellapak_250Y', col_type='abs')
        self.transport.main_model = self.main_model
        self.subblocks['transport'] = self.transport
        self.transport.main_model = self.main_model
        blk.trans = Block(rule=self.transport)
        self.transport.pyomo_block = blk.trans
        
    def _chem_eq(self, blk):
        """
        Add the chemical equilibruim constraints
        """
        p_henry_co2 = self.main_model.p_henry_co2
        p_V_liq = self.main_model.p_V_liq
        p_vapp = self.main_model.p_vapp
        p_logK = self.main_model.p_logKeq
        # Some aliases
        y_star = blk.y_star
        logC = blk.logC
        C = blk.C
        # Constraint sets
        blk.eq_y_star = Constraint(blk.z, blk.lcomp, noruleinit=True)
        blk.eq_MEA_bal = Constraint(blk.z, noruleinit=True)
        blk.eq_CO2_bal = Constraint(blk.z, noruleinit=True)
        blk.eq_charge_bal = Constraint(blk.z, noruleinit=True)
        blk.eq_rxn = Constraint(blk.z, blk.rxns, noruleinit=True)
        blk.eq_C = Constraint(blk.z, blk.ccomp, noruleinit=True)
        y = blk.y_l
        # Constraint equations
        for z in blk.z:
            # Grab some nicer names for properties.
            P = blk.P[z]
            T = blk.T_l[z]
            # C to logC equations
            for i in blk.ccomp:
                blk.eq_C.add((z,i), expr=C[z,i] == exp(logC[z,i]))
            # Balance equations
            blk.eq_MEA_bal.add(
                z, expr=0 == -y[z,'MEA']/p_V_liq(T, y[z,'CO2'], y[z,'H2O'], 0.0, y[z,'MEA']) +
                C[z,'RNH2'] + C[z,'RNH3+'] + C[z,'RNHCOO-'])
            blk.eq_CO2_bal.add(
                z, expr=0 == -y[z,'CO2']/p_V_liq(T, y[z,'CO2'], y[z,'H2O'], 0.0, y[z,'MEA']) + 
                C[z,'CO2'] + C[z,'HCO3-'] + C[z,'CO3--'] + C[z,'RNHCOO-'])
            blk.eq_charge_bal.add(
                z, expr=0 == C[z,'RNH3+'] + C[z,'H3O+'] 
                - C[z,'HCO3-'] - C[z,'OH-'] - 2*C[z,'CO3--'] - C[z,'RNHCOO-'])
            # Chemical Equilibrium
            blk.eq_rxn.add(
                (z,1), expr=0 == 
                logC[z,'OH-'] + logC[z,'H3O+'] - p_logK(1, T))
            blk.eq_rxn.add(
                (z,2), expr=0 == 
                logC[z,'HCO3-'] + logC[z,'H3O+'] - 
                logC[z,'CO2'] - p_logK(2, T))
            blk.eq_rxn.add(
                (z,3), expr=0 == 
                logC[z,'CO3--'] + logC[z,'H3O+'] -
                logC[z,'HCO3-'] - p_logK(3, T))
            blk.eq_rxn.add(
                (z,8), expr=0 == 
                logC[z,'RNH2'] + logC[z,'HCO3-'] -
                logC[z,'RNHCOO-'] - p_logK(8, T))
            blk.eq_rxn.add(
                (z,9), expr=0 == 
                logC[z,'RNH2'] + logC[z,'H3O+'] -
                logC[z,'RNH3+'] - p_logK(9, T))
                
            blk.eq_y_star.add(
                (z, 'H2O'), expr=0 == 
                y_star[z,'H2O']*P - p_vapp(1, T, y[z,'H2O']))
            blk.eq_y_star.add(
                (z, 'MEA'), expr=0 == 
                y_star[z,'MEA']*P - p_vapp(3, T, y[z,'MEA']))
            blk.eq_y_star.add(
                (z, 'CO2'), expr = 0 == 
                y_star[z,'CO2']*P - 
                p_henry_co2(T, y[z,'H2O'], y[z,'MEA'])*C[z, 'CO2'])
        
    def _mass_balance(self, blk):
        """
        Add the mass balance constriants.
        """
        p_V_vap = self.main_model.p_V_vap
        p_V_liq = self.main_model.p_V_liq
        # Get some shorter variable names to add balance constraints
        L = blk.L
        Cu_v = blk.Cu_v
        Cu_l = blk.Cu_l
        N_v = blk.N_v
        ul = blk.ul
        uv = blk.uv
        y_star = blk.y_star
        dy = blk.dy
        # Derivatives, shorter names
        dCu_ldz = blk.dCu_ldz
        dCu_vdz = blk.dCu_vdz
        # Transport shorter names
        K_v = blk.trans.K_v
        ae = blk.trans.ae
        eps = blk.trans.eps
        eps_v = blk.trans.eps_v
        eps_l = blk.trans.eps_l
        T_v = blk.T_v
        
        # A few simple constraints for flow rates
        blk.eq_A = Constraint(expr=0 == blk.A - 3.14159*(blk.D/2.0)**2)
        blk.eq_F_vap_in = Constraint(
            expr=0 == 
            blk.F_vap_in - blk.A*sum(blk.Cu_v[1,i] for i in blk.comp))
        blk.eq_F_vap_out = Constraint(
            expr=0 == 
            blk.F_vap_out - blk.A*sum(blk.Cu_v[0,i] for i in blk.comp))
        blk.eq_F_liq_in = Constraint(
            expr=0 == 
            blk.F_liq_in - blk.A*sum(blk.Cu_l[0,i] for i in blk.lcomp))
        blk.eq_F_liq_out = Constraint(
            expr=0 == 
            blk.F_liq_out - blk.A*sum(blk.Cu_l[1,i] for i in blk.lcomp))
        # calculate compositions for easier use with inlet and outlet
        # ports and specifications.
        blk.eq_y_vap1 = Constraint(blk.comp, noruleinit=True)
        blk.eq_y_vap2 = Constraint(blk.comp, noruleinit=True)
        blk.eq_y_liq1 = Constraint(blk.lcomp, noruleinit=True)
        blk.eq_y_liq2 = Constraint(blk.lcomp, noruleinit=True)
        for i in blk.comp:
            blk.eq_y_vap1.add(
                i, expr=0 == -blk.y_vap1[i] + blk.y_v[1,i])
            blk.eq_y_vap2.add(
                i, expr=0 == -blk.y_vap2[i] + blk.y_v[0,i])
        for i in blk.lcomp:
            blk.eq_y_liq1.add(
                i, expr=0 == -blk.y_liq1[i] + blk.y_l[0,i])
            blk.eq_y_liq2.add(
                i, expr=0 == -blk.y_liq2[i] + blk.y_l[1,i])
                
        # Add constraints with equations over each discrete point.
        blk.eq_dy = Constraint(blk.z, blk.lcomp, noruleinit=True)
        blk.eq_Nv = Constraint(blk.z, blk.lcomp, noruleinit=True)
        blk.eq_Nvtot = Constraint(blk.z, noruleinit=True)
        blk.eq_sum_frac_l = Constraint(blk.z, noruleinit=True)
        blk.eq_sum_frac_v = Constraint(blk.z, noruleinit=True)
        blk.eq_Cu_l = Constraint(blk.z, blk.lcomp, noruleinit=True)
        blk.eq_Cu_v = Constraint(blk.z, blk.comp, noruleinit=True)
        blk.eq_mass_l = Constraint(blk.z, blk.lcomp, noruleinit=True)
        blk.eq_mass_v = Constraint(blk.z, blk.lcomp, noruleinit=True)
        
        # Equations
        pn = len(blk.z) #number of points/last point
        for pi0, z in enumerate(blk.z):
            pi = pi0 + 1 # pi is the point index starting at 1 and z is 
                         # dimentionless distance
            if z > 0: zb = blk.z[pi-1] # get key for point lower
            if z < 1: zf = blk.z[pi+1] # get key for point higher

            # Sum of mole fraction in both phases = 1, skipping 
            # inlet, assuming those are properly specified elsewhere
            if pi != pn: 
                blk.eq_sum_frac_v.add(
                    z, expr=0 == 1 - 
                    sum(blk.y_v[z,i] for i in blk.comp))
            
            if pi != 1: 
                blk.eq_sum_frac_l.add(
                    z, expr=0 == 1 - 
                    sum(blk.y_l[z,i] for i in blk.lcomp))
            P = blk.P[z]
            # Mass balance terms
            y = blk.y_l
            T_l = blk.T_l[z]
            blk.eq_Nvtot.add(z, expr=blk.N_vtot[z] == sum(N_v[z,i] for i in blk.lcomp))
            for i in blk.lcomp:
                y_i = blk.y_v[z,i]
                x_i = blk.y_l[z,i]
                blk.eq_dy.add(
                    (z,i), expr=0 == -dy[z,i] + y_i - y_star[zf,i])
                blk.eq_Nv.add(
                    (z,i), expr=0 == -N_v[z,i] + 
                    K_v[z,i]*ae[z]*dy[z,i]/p_V_vap(T_v[z], blk.P[z]) + blk.N_vtot[z]*y_i)
                blk.eq_Cu_v.add(
                    (z,i), expr=0 == -Cu_v[z,i] + 
                        y_i*uv[z]/p_V_vap(T_v[z], blk.P[z]))
                blk.eq_Cu_l.add(
                    (z,i), expr=0 == -Cu_l[z,i]*p_V_liq(T_l, y[z,'CO2'], y[z,'H2O'], 0.0, y[z,'MEA']) + x_i*ul[z])
            # Vapor phase N2 is a bit different assumed is not in liquid
            y_i = blk.y_v[z,'N2']
            blk.eq_Cu_v.add(
                (z,'N2'), expr=0 == -Cu_v[z,'N2'] + y_i*uv[z]/p_V_vap(T_v[z], blk.P[z]))
            
            # Mass balance equations
            if z < 1: # N2 special case
                blk.eq_mass_v.add(
                    (z,'N2'), expr=0 == dCu_vdz[z,'N2'])
            
            # Rest of components vapor and liquid
            for i in blk.lcomp:
                if z < 1: 
                    blk.eq_mass_v.add(
                        (z,i), expr=0 == 
                        dCu_vdz[z,i] - blk.init_mt_scale*L*N_v[z,i])
                if z > 0: 
                    blk.eq_mass_l.add(
                        (z,i), expr=0 == 
                        dCu_ldz[z,i] - blk.init_mt_scale*L*N_v[zb,i])
        
    def _energy_balance(self, blk):
        """
        Add the energy balance constraints.
        """
        # Ideal gas heat capacity function
        p_cp = self.main_model.p_cp_ig
        p_V_vap = self.main_model.p_V_vap
        p_V_liq = self.main_model.p_V_liq
        p_cp_liq = self.main_model.p_cp_liq
        p_hvap_H2O = self.main_model.p_hvap_H2O
        p_hvap_MEA = self.main_model.p_hvap_MEA
        p_habs_CO2 = self.main_model.p_habs_CO2
        # Get some shorter variable names to add balance constraints
        L = blk.L
        ul = blk.ul
        uv = blk.uv
        T_v = blk.T_v
        T_l = blk.T_l
        P = blk.P
        N_v = blk.N_v
        deltaT = blk.deltaT
        Q_v = blk.Q_v
        # Derivatives, shorter names
        dT_vdz = blk.dT_vdz
        dT_ldz = blk.dT_ldz
        # Transport, shorter names
        ae = blk.trans.ae
        eps = blk.trans.eps
        eps_v = blk.trans.eps_v
        eps_l = blk.trans.eps_l
        h_vl = blk.trans.h_vl

        # Add some constraint sets
        blk.eq_Qv = Constraint(blk.z, noruleinit=True)
        blk.eq_h_abs_vap = Constraint(blk.z, noruleinit=True)
        #blk.eq_T_v = Constraint(blk.z, noruleinit=True)
        #blk.eq_T_l = Constraint(blk.z, noruleinit=True)
        blk.eq_deltaT = Constraint(blk.z, noruleinit=True)
        blk.eq_energy_l = Constraint(blk.z, noruleinit=True)
        blk.eq_energy_v = Constraint(blk.z, noruleinit=True)  
        
        # Constriant equations
        for pi0, z in enumerate(blk.z):
            pi = pi0 + 1 # z is dimensionless length, pi integer index 
            
            #Define a forward and backward point
            if z > 0: 
                zb = blk.z[pi-1]
            if z < 1: 
                zf = blk.z[pi+1]
            # Energy balance equations
                          
            if z > 0:
                y = blk.y_l 
                # H sum(h_vap) + sum(h_abs) for liquid at z
                blk.eq_h_abs_vap.add(
                    z, expr=blk.h_abs_vap[z] == 
                    p_habs_CO2(T_l[zb])*N_v[zb, 'CO2'] +
                    p_hvap_H2O(T_l[zb])*N_v[zb, 'H2O'] +
                    p_hvap_MEA(T_l[zb])*N_v[zb, 'MEA'])
                #Liquid energy balance
                blk.eq_energy_l.add(
                    z, expr=0 == 
                    -p_cp_liq(T_l[z], y[z,'H2O'], y[z, 'MEA'])*
                    ul[z]*dT_ldz[z]/
                    p_V_liq(T_l[z], y[z,'CO2'], y[z,'H2O'], 0.0, y[z,'MEA'])
                    - L*Q_v[zb]
                    + L*blk.h_abs_vap[z]*blk.init_ht_scale)
            if z < 1:
                y = blk.y_v
                # delta T between phases
                blk.eq_deltaT.add(
                    z, expr = 0 == -deltaT[z] + T_l[zf] - T_v[z])
                # Heat Transfer Between pahses
                blk.eq_Qv.add(
                    z, expr =0== -Q_v[z] + ae[z]*h_vl[z]*deltaT[z])
                # Vapor energy balance
                blk.eq_energy_v.add(
                    z, expr=0 == 
                    p_cp(T_v[z], y[z,'CO2'], y[z,'H2O'], y[z, 'N2'], y[z, 'MEA'])\
                    *uv[z]*dT_vdz[z]/p_V_vap(T_v[z], blk.P[z]) + 
                    blk.init_ht_scale*L*Q_v[z])
                    

            
    def _default_init(self, model, solver='ipopt', m_steps=[1.0],
                      h_steps = [0.1, 0.2, 0.5, 1.0]):
        """
        Initialize the absorber model.  This has a lot of steps and 
        takes a while, so it prints output to let you know what's going
        on.
        
        Once the intialization is done the state of the models is:
        P in all property blocks is fixed
        T at the vapor and liquid inlets is fixed
        y at the vapor and liquid inlets are fixed
        ul as liquid inlet is fixed
        uv at vapor inlet is fixed
        
        Args:
        model: for now this take the pyomo model. need to fix this so
               it only solves the absorber, but havent worked out the
               details of the best way to initialize blocks
        """
        # Start time for total elapsed time
        total_time_start = time.time()
        
        # Create a solver to solve initialization steps
        opt = SolverFactory(solver)
        stee = False
        pyutilib.services.TempfileManager.tempdir = './'
        #opt.options['halt_on_ampl_error'] = 'yes'
        if solver=='ipopt':
            #opt.options['linear_solver'] = 'ma57'
            #opt.options['linear_system_scaling'] = 'mc19'
            opt.options['mu_init'] = '1e-8'
            opt.options['bound_push'] = '1e-8'
        elif solver=='petsc':
            opt.options['-snes_monitor'] = ''
            opt.options['-snes_max_it'] = '20000'
            opt.options['-snes_max_funcs'] = '60000'
            opt.options['-snes_atol'] = '1e-10'
            opt.options['-snes_rtol'] = '1e-8'
            #opt.options['-ksp_monitor'] = ''
            #opt.options['-use_bounds'] = ''
            #opt.options['-snes_type'] = 'vinewtonssls'
            #opt.options['-remove_eq'] = ''
            #opt.options['-scale_eqs'] = '0'
            #opt.options['-scale_vars'] = '0'
        else:
            #other solvers
            pass
            
        print("\n")
        print("-------------------------------------------------------")
        print("Absorber initialization\n")
        # Fix the some variables for initial property and transport steps
        self.pyomo_block.L.fix()
        self.pyomo_block.D.fix()
        self.pyomo_block.ul.fix()
        self.pyomo_block.uv.fix()
        self.pyomo_block.P.fix()
        self.pyomo_block.T_l.fix()
        self.pyomo_block.T_v.fix()
        self.pyomo_block.y_l.fix()
        self.pyomo_block.y_v.fix()
        # Deactiavte some constraints
        self.pyomo_block.eq_energy_l.deactivate()
        self.pyomo_block.eq_energy_v.deactivate()
        self.pyomo_block.eq_h_abs_vap.deactivate()
        self.pyomo_block.eq_mass_l.deactivate()
        self.pyomo_block.eq_mass_v.deactivate()
        self.pyomo_block.eq_sum_frac_l.deactivate()
        self.pyomo_block.eq_sum_frac_v.deactivate()
        self.pyomo_block.eq_Nv.deactivate()
        self.pyomo_block.eq_Qv.deactivate()
        self.pyomo_block.trans.eq_h_vl.deactivate()
        self.pyomo_block.trans.eq_k_l_CO2.deactivate()
        self.pyomo_block.trans.eq_h_vl.deactivate()
        self.pyomo_block.trans.eq_k_v.deactivate()  
        self.pyomo_block.trans.eq_K_v.deactivate()  
        self.pyomo_block.trans.eq_E.deactivate()  
        self.pyomo_block.trans.eq_kap.deactivate()
        # Set initialization parameters to zero, so can gradually
        # turn on terms in the mass and energy balance
        self.pyomo_block.init_ht_scale = 0
        self.pyomo_block.init_mt_scale = 0
        ts = time.time()
        print("Step 1, Initialize Properties, Time: ", end="")
        results = opt.solve(model)
        print("{}, {}".format(model_util.hhmmss(time.time() - ts), results.solver.message))

        #print("Step 2, Mass balance d/dz=0, Time: ", end="")
        self.pyomo_block.ul.unfix()
        self.pyomo_block.uv.unfix()
        self.pyomo_block.ul[0].fix()
        self.pyomo_block.uv[1].fix()
        self.pyomo_block.eq_mass_l.activate()
        self.pyomo_block.eq_mass_v.activate()
        self.pyomo_block.eq_sum_frac_l.activate()
        self.pyomo_block.eq_sum_frac_v.activate()
        for k,v in self.pyomo_block.y_l.iteritems():
            if k[0] != 0: # this is a bit lazy but is easy
                v.unfix()
        for k,v in self.pyomo_block.y_v.iteritems():
            if k[0] != 1:
                v.unfix()      
        #results = opt.solve(model)
        #print(results)
        #print(model_util.hhmmss(time.time() - ts))

        #print("Step 3, Time: ", end="")
        self.pyomo_block.eq_energy_l.activate()
        self.pyomo_block.eq_energy_v.activate()
        self.pyomo_block.eq_h_abs_vap.activate()
        self.pyomo_block.T_l.unfix()
        self.pyomo_block.T_v.unfix()
        self.pyomo_block.Q_v.fix(0.0)
        self.pyomo_block.T_l[0].fix()
        self.pyomo_block.T_v[1].fix()
        ts = time.time()
        #results = opt.solve(model)
        #print("{}, {}".format(model_util.hhmmss(time.time() - ts), results.solver.message))
        
        print("Step 4, Time: ", end="")
        self.pyomo_block.trans.eq_kap.activate()
        self.pyomo_block.trans.eq_k_v.activate()
        self.pyomo_block.trans.eq_k_l_CO2.activate()
        self.pyomo_block.trans.eq_E.activate()
        self.pyomo_block.trans.eq_K_v.activate()
        model.absorb.eq_Nv.activate()
        model.absorb.trans.eq_h_vl.activate() 
        model.absorb.Q_v.unfix()
        model.absorb.eq_Qv.activate()
        ts = time.time()
        results = opt.solve(model, tee=stee)
        print("{}, {}".format(model_util.hhmmss(time.time() - ts), results.solver.message))
        
        print("Step 5, turn up the mass and heat transfer in steps")
        for i in m_steps:
            print("Step 5a mass -- {0}, Time: ".format(i), end="")
            model.absorb.init_mt_scale = i
            ts = time.time()
            results = opt.solve(model, tee=stee)
            print("{}, {}".format(model_util.hhmmss(time.time() - ts), results.solver.message))
        for i in h_steps:
            #print(results)
            print("Step 5b heat -- {0}, Time: ".format(i), end="")
            model.absorb.init_ht_scale = i
            ts = time.time()
            #results = opt.solve(model, tee=stee, symbolic_solver_labels=True)
            results = opt.solve(model, tee=stee)
            print("{}, {}".format(model_util.hhmmss(time.time() - ts), results.solver.message))
            
        print("\n\nTotal initialization time: {0}".format(
            model_util.hhmmss(time.time()-total_time_start)))
        print("Initialization complete")
        print("-------------------------------------------------------")
        
        
        
        
        
        
        
        
        
