##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for framework example.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import SolverFactory, value
from pyomo.opt import SolverStatus, TerminationCondition
from idaes.ui.report import degrees_of_freedom

# Import file to be tested
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'framework_example.py')
fe = import_file(example)


@pytest.fixture(scope="module")
def model_object():
    """Returns an instance of the model for testing.
    """
    m = fe.build_model()
    return m


# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


def test_build_model(model_object):
    # Try constructing the model and do a basic check
    assert value(model_object.fs_obj.C01.work_isentropic[0]) == 0.0
    assert value(model_object.fs_obj.HX02.heat_transfer_area) == 1.0


def test_setInputs(model_object):
    m = model_object

    for t in m.fs_obj.time:
        # ---------------------------------------------------------------------
        # Unit C01
        m.fs_obj.C01.inlet[t, "1"].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
        m.fs_obj.C01.inlet[t, "1"].vars["temperature"].fix(866.5)
        m.fs_obj.C01.inlet[t, "1"].vars["pressure"].fix(2.97e7)

        m.fs_obj.C01.inlet[t, "2"].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
        m.fs_obj.C01.inlet[t, "2"].vars["temperature"].fix(866.5)
        m.fs_obj.C01.inlet[t, "2"].vars["pressure"].fix(2.97e7)

        m.fs_obj.C01.volume.fix(1.0)
        m.fs_obj.C01.deltaP.fix(-1e7)
        m.fs_obj.C01.efficiency_isentropic.fix(0.83)

        m.fs_obj.C01.holdup.material_accumulation[0, "Vap", "H2O"].fix(0.0)
        m.fs_obj.C01.holdup.energy_accumulation[0, "Vap"].fix(0.0)

        # ---------------------------------------------------------------------
        # Unit HX02
        m.fs_obj.HX02.side_2_inlet[t].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
        m.fs_obj.HX02.side_2_inlet[t].vars["temperature"].fix(500.0)
        m.fs_obj.HX02.side_2_inlet[t].vars["pressure"].fix(2.97e7)

        m.fs_obj.HX02.volume_side_1.fix(1.0)
        m.fs_obj.HX02.volume_side_2.fix(1.0)
        m.fs_obj.HX02.heat_transfer_coefficient.fix(5.0)
        m.fs_obj.HX02.heat_transfer_area.fix(100000.0)

        m.fs_obj.HX02.side_1.outlet_splitter.split_fraction[t, "1"].fix(0.4)

        # Initial Condition - Steady-State
        m.fs_obj.HX02.side_1.material_accumulation[0, "Vap", "H2O"].fix(0.0)
        m.fs_obj.HX02.side_1.energy_accumulation[0, "Vap"].fix(0.0)
        m.fs_obj.HX02.side_2.material_accumulation[0, "Vap", "H2O"].fix(0.0)
        m.fs_obj.HX02.side_2.energy_accumulation[0, "Vap"].fix(0.0)

    # Check values were assigned properly
    assert degrees_of_freedom(model_object) == 0


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization(model_object):
    results = fe.initialize_model(model_object, solver, stee=True)

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_solution(model_object):
    results = fe.simulation1(model_object, solver, stee=True)

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    # Check the solution numbers
    assert (pytest.approx(-273676183.15, abs=1e-3) ==
            value(model_object.fs_obj.C01.work_mechanical[
                    model_object.fs_obj.time.last()]))
    assert (pytest.approx(-110885237.928, abs=1e-3) ==
            value(model_object.fs_obj.HX02.heat[
                    model_object.fs_obj.time.last()]))
