##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for ports.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, Expression, Set, Var
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        UnitBlockData, Holdup0D, Holdup1D, HoldupStatic, \
                        PropertyParameterBase, PropertyBlockDataBase
from idaes.core.ports import Port, InletMixer, InletMixerData, \
                             OutletSplitter, OutletSplitterData
from pyomo.common.config import ConfigValue


# -----------------------------------------------------------------------------
# Setup a basic unit model for testing
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("Unit")
class UnitData(UnitBlockData):
    def build(self):
        pass


@declare_process_block_class("PropertyParameterBlock")
class PropertyParameterData(PropertyParameterBase):
    def build(self):
        super(PropertyParameterData, self).build()

    def get_supported_properties(self):
        return {'flow': {'method': None, 'units': 'mol/s'},
                'pressure': {'method': None, 'units': 'Pa'},
                'temperature': {'method': None, 'units': 'K'}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


@declare_process_block_class("PropertyBlock")
class PropertyBlockData(PropertyBlockDataBase):
    def build(self):
        super(PropertyBlockData, self).build()
        self.phase_list = Set(initialize=["phase1", "phase2"])
        self.component_list = Set(initialize=["comp1", "comp2"])
        self.pressure = Var(initialize=1.0)
        self.temperature = Var(initialize=2.0)
        self.flow = Var(self.component_list, initialize=3.0)

        def material_terms(b, p, j):
            return 1
        self.material_balance_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_terms)

        def energy_terms(b, p):
            return 2
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_terms)

        def material_density_terms(b, p, j):
            return 3
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_terms)

        def energy_density_terms(b, p):
            return 4
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_terms)

    def declare_port_members(b):
        members = {"a": b.pressure,
                   "b": b.temperature,
                   "c": b.flow}
        return members


@pytest.fixture()
def m0():
    """
    Build a basic Unit model with one Holdup for testing.
    """
    m0 = ConcreteModel()
    m0.fs = Flowsheet(dynamic=False)
    m0.fs.prop = PropertyParameterBlock()
    m0.fs.config.default_property_package = m0.fs.prop

    m0.fs.u = Unit()
    m0.fs.u.config.declare("property_package", ConfigValue(default=m0.fs.prop))
    m0.fs.u.config.declare("property_package_args", ConfigValue(default={}))
    m0.fs.u._setup_dynamics()
    m0.fs.u.holdup = Holdup0D()
    return m0


@pytest.fixture()
def m1f():
    """
    Build a basic Unit model with one Holdup for testing.
    """
    m1f = ConcreteModel()
    m1f.fs = Flowsheet(dynamic=False)
    m1f.fs.prop = PropertyParameterBlock()
    m1f.fs.config.default_property_package = m1f.fs.prop

    m1f.fs.u = Unit()
    m1f.fs.u.config.declare("property_package",
                            ConfigValue(default=m1f.fs.prop))
    m1f.fs.u.config.declare("property_package_args", ConfigValue(default={}))
    m1f.fs.u._setup_dynamics()
    m1f.fs.u.holdup = Holdup1D()
    return m1f


@pytest.fixture()
def m1b():
    """
    Build a basic Unit model with one Holdup for testing.
    """
    m1b = ConcreteModel()
    m1b.fs = Flowsheet(dynamic=False)
    m1b.fs.prop = PropertyParameterBlock()
    m1b.fs.config.default_property_package = m1b.fs.prop

    m1b.fs.u = Unit()
    m1b.fs.u.config.declare("property_package",
                            ConfigValue(default=m1b.fs.prop))
    m1b.fs.u.config.declare("property_package_args", ConfigValue(default={}))
    m1b.fs.u._setup_dynamics()
    m1b.fs.u.holdup = Holdup1D(flow_direction="backward")
    return m1b


@pytest.fixture()
def ms():
    """
    Build a basic Unit model with one Holdup for testing.
    """
    ms = ConcreteModel()
    ms.fs = Flowsheet(dynamic=False)
    ms.fs.prop = PropertyParameterBlock()
    ms.fs.config.default_property_package = ms.fs.prop

    ms.fs.u = Unit()
    ms.fs.u.config.declare("property_package", ConfigValue(default=ms.fs.prop))
    ms.fs.u.config.declare("property_package_args", ConfigValue(default={}))
    ms.fs.u._setup_dynamics()
    ms.fs.u.holdup = HoldupStatic()
    return ms


# -----------------------------------------------------------------------------
# Test common Port class
@declare_process_block_class("PortFrame")
class PortFrameData(Port):
    def build(self):
        pass


def test_Port_config(m0):
    # Test that config arguments take correct values
    m0.fs.u.holdup.port = PortFrame()

    m0.fs.u.holdup.port.config.has_material_balance = False
    m0.fs.u.holdup.port.config.has_energy_balance = False
    m0.fs.u.holdup.port.config.has_momentum_balance = False

    m0.fs.u.holdup.port.config.has_material_balance = True
    m0.fs.u.holdup.port.config.has_energy_balance = True
    m0.fs.u.holdup.port.config.has_momentum_balance = True

    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.has_material_balance = 'foo'
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.has_material_balance = 10
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.has_energy_balance = 'foo'
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.has_energy_balance = 10
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.has_momentum_balance = 'foo'
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.has_momentum_balance = 10


def test_make_common_references(m0):
    m0.fs.u.holdup.port = PortFrame()
    m0.fs.u.holdup.port._make_common_references()

    assert hasattr(m0.fs.u.holdup.port, "time")
    assert hasattr(m0.fs.u.holdup.port, "component_list")
    assert hasattr(m0.fs.u.holdup.port, "phase_list")


def test_get_build_flags(m0):
    m0.fs.u.holdup.port = PortFrame()

    m0.fs.u.holdup.config.material_balance_type = 'none'
    m0.fs.u.holdup.config.energy_balance_type = 'none'
    m0.fs.u.holdup.config.momentum_balance_type = 'none'
    m0.fs.u.holdup.port._get_build_flags()

    assert m0.fs.u.holdup.port.config.has_material_balance is False
    assert m0.fs.u.holdup.port.config.has_energy_balance is False
    assert m0.fs.u.holdup.port.config.has_momentum_balance is False

    m0.fs.u.holdup.config.material_balance_type = 'component_phase'
    m0.fs.u.holdup.config.energy_balance_type = 'total'
    m0.fs.u.holdup.config.momentum_balance_type = 'total'
    m0.fs.u.holdup.port._get_build_flags()

    assert m0.fs.u.holdup.port.config.has_material_balance
    assert m0.fs.u.holdup.port.config.has_energy_balance
    assert m0.fs.u.holdup.port.config.has_momentum_balance


def test_Port_build(m0):
    m0.fs.u.holdup.port = PortFrame()
    super(PortFrameData, m0.fs.u.holdup.port).build()

    assert hasattr(m0.fs.u.holdup.port, "time")
    assert hasattr(m0.fs.u.holdup.port, "component_list")
    assert hasattr(m0.fs.u.holdup.port, "phase_list")
    assert m0.fs.u.holdup.port.config.has_material_balance
    assert m0.fs.u.holdup.port.config.has_energy_balance
    assert m0.fs.u.holdup.port.config.has_momentum_balance


# -----------------------------------------------------------------------------
# Test InletMixer
@declare_process_block_class("InletMixerFrame")
class InletMixerFrameData(InletMixerData):
    def build(self):
        super(InletMixerData, self).build()
        pass


def test_InletMixer_make_references_0D(m0):
    m0.fs.u.holdup.port = InletMixerFrame()
    m0.fs.u.holdup.port._make_references()

    assert (m0.fs.u.holdup.port.properties_in ==
            m0.fs.u.holdup.properties_in)


def test_InletMixer_make_references_1D_forward(m1f):
    m1f.fs.u.holdup.port = InletMixerFrame()
    m1f.fs.u.holdup.port._make_references()

    for t in m1f.fs.time:
        assert (m1f.fs.u.holdup.port.properties_in[t] ==
                m1f.fs.u.holdup.properties[t, 0])


def test_InletMixer_make_references_1D_backward(m1b):
    m1b.fs.u.holdup.port = InletMixerFrame()
    m1b.fs.u.holdup.port._make_references()

    for t in m1b.fs.time:
        assert (m1b.fs.u.holdup.port.properties_in[t] ==
                m1b.fs.u.holdup.properties[t, 1])


def test_InletMixer_make_references_static(ms):
    ms.fs.u.holdup.port = InletMixerFrame()
    ms.fs.u.holdup.port._make_references()

    for t in ms.fs.time:
        assert (ms.fs.u.holdup.port.properties_in[t] ==
                ms.fs.u.holdup.properties[t])


def test_InletMixer_build_property_blocks(m0):
    m0.fs.u.holdup.port = InletMixerFrame()
    m0.fs.u.holdup.port.config.inlets = ["1", "2"]
    m0.fs.u.holdup.port._build_property_blocks()

    assert hasattr(m0.fs.u.holdup.port, "inlet_idx")
    assert hasattr(m0.fs.u.holdup.port, "properties")
    for k in m0.fs.u.holdup.port.properties.keys():
        assert k in [(0, "1"), (0, "2")]


def test_InletMixer_mix_material(m0):
    m0.fs.u.holdup.port = InletMixerFrame()
    m0.fs.u.holdup.port.config.inlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port._mix_material()

    assert hasattr(m0.fs.u.holdup.port, "material_mixing")


def test_InletMixer_mix_energy(m0):
    m0.fs.u.holdup.port = InletMixerFrame()
    m0.fs.u.holdup.port.config.inlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port._mix_energy()

    assert hasattr(m0.fs.u.holdup.port, "energy_mixing")


def test_InletMixer_mix_pressure(m0):
    m0.fs.u.holdup.port = InletMixerFrame()
    m0.fs.u.holdup.port.config.inlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port._mix_pressure()

    assert hasattr(m0.fs.u.holdup.port, "inlet_pressure")
    assert hasattr(m0.fs.u.holdup.port, "eps_pressure")
    assert hasattr(m0.fs.u.holdup.port, "minimum_pressure")
    assert hasattr(m0.fs.u.holdup.port, "minimum_pressure_constraint")


def test_InletMixer_build(m0):
    m0.fs.u.holdup.port = InletMixer(inlets=["1", "2"])

    assert hasattr(m0.fs.u.holdup.port, "inlet_idx")
    assert hasattr(m0.fs.u.holdup.port, "properties")
    for k in m0.fs.u.holdup.port.properties.keys():
        assert k in [(0, "1"), (0, "2")]
    assert hasattr(m0.fs.u.holdup.port, "material_mixing")
    assert hasattr(m0.fs.u.holdup.port, "energy_mixing")
    assert hasattr(m0.fs.u.holdup.port, "inlet_pressure")
    assert hasattr(m0.fs.u.holdup.port, "eps_pressure")
    assert hasattr(m0.fs.u.holdup.port, "minimum_pressure")
    assert hasattr(m0.fs.u.holdup.port, "minimum_pressure_constraint")


def test_InletMixer_build_no_inlets_arg(m0):
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port = InletMixer()


# -----------------------------------------------------------------------------
# Test OutletMixer
@declare_process_block_class("OutletSplitterFrame")
class OutletSplitterFrameData(OutletSplitterData):
    def build(self):
        super(OutletSplitterData, self).build()
        pass


def test_OutletSplitter_config(m0):
    # Check that split_type takes correct arguments
    m0.fs.u.holdup.port = OutletSplitterFrame()

    m0.fs.u.holdup.port.config.split_type = "flow"
    m0.fs.u.holdup.port.config.split_type = "phase"
    m0.fs.u.holdup.port.config.split_type = "component"
    m0.fs.u.holdup.port.config.split_type = "total"

    # Check that invalid arguments are rejected
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.split_type = "foo"  # invalid str
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.split_type = 1  # invalid int
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.split_type = 1.0  # invalid float
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.split_type = [1]  # invalid list
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.split_type = {"1": 1}  # invalid dict
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port.config.split_type = m0.fs.prop  # invalid object


def test_OutletSplitter_make_references_0D(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port._make_references()

    assert (m0.fs.u.holdup.port.properties_out ==
            m0.fs.u.holdup.properties_out)


def test_OutletSplitter_make_references_1D_forward(m1f):
    m1f.fs.u.holdup.port = OutletSplitterFrame()
    m1f.fs.u.holdup.port._make_references()

    for t in m1f.fs.time:
        assert (m1f.fs.u.holdup.port.properties_out[t] ==
                m1f.fs.u.holdup.properties[t, 1])


def test_OutletSplitter_make_references_1D_backward(m1b):
    m1b.fs.u.holdup.port = OutletSplitterFrame()
    m1b.fs.u.holdup.port._make_references()

    for t in m1b.fs.time:
        assert (m1b.fs.u.holdup.port.properties_out[t] ==
                m1b.fs.u.holdup.properties[t, 0])


def test_OutletSplitter_make_references_static(ms):
    ms.fs.u.holdup.port = OutletSplitterFrame()
    ms.fs.u.holdup.port._make_references()

    for t in ms.fs.time:
        assert (ms.fs.u.holdup.port.properties_out[t] ==
                ms.fs.u.holdup.properties[t])


def test_OutletSplitter_build_property_blocks(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._build_property_blocks()

    assert hasattr(m0.fs.u.holdup.port, "outlet_idx")
    assert hasattr(m0.fs.u.holdup.port, "properties")
    for k in m0.fs.u.holdup.port.properties.keys():
        assert k in [(0, "1"), (0, "2")]


def test_OutletSplitter_make_split_fraction_flow(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "flow"
    m0.fs.u.holdup.port._make_split_fraction()

    assert hasattr(m0.fs.u.holdup.port, "split_fraction")
    for k in m0.fs.u.holdup.port.split_fraction.keys():
        assert k in [(0, "1"), (0, "2")]
    assert hasattr(m0.fs.u.holdup.port, "sum_split_fractions")
    for k in m0.fs.u.holdup.port.sum_split_fractions.keys():
        assert k in [0]


def test_OutletSplitter_make_split_fraction_phase(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "phase"
    m0.fs.u.holdup.port._make_split_fraction()

    assert hasattr(m0.fs.u.holdup.port, "split_fraction")
    for k in m0.fs.u.holdup.port.split_fraction.keys():
        assert k in [(0, "1", "phase1"), (0, "1", "phase2"),
                     (0, "2", "phase1"), (0, "2", "phase2")]
    assert hasattr(m0.fs.u.holdup.port, "sum_split_fractions")
    for k in m0.fs.u.holdup.port.sum_split_fractions.keys():
        assert k in [(0, "phase1"), (0, "phase2")]


def test_OutletSplitter_make_split_fraction_component(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "component"
    m0.fs.u.holdup.port._make_split_fraction()

    assert hasattr(m0.fs.u.holdup.port, "split_fraction")
    for k in m0.fs.u.holdup.port.split_fraction.keys():
        assert k in [(0, "1", "comp1"), (0, "1", "comp2"),
                     (0, "2", "comp1"), (0, "2", "comp2")]
    assert hasattr(m0.fs.u.holdup.port, "sum_split_fractions")
    for k in m0.fs.u.holdup.port.sum_split_fractions.keys():
        assert k in [(0, "comp1"), (0, "comp2")]


def test_OutletSplitter_make_split_fraction_total(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "total"
    m0.fs.u.holdup.port._make_split_fraction()

    assert hasattr(m0.fs.u.holdup.port, "split_fraction")
    for k in m0.fs.u.holdup.port.split_fraction.keys():
        assert k in [(0, "1", "phase1", "comp1"), (0, "1", "phase1", "comp2"),
                     (0, "1", "phase2", "comp1"), (0, "1", "phase2", "comp2"),
                     (0, "2", "phase1", "comp1"), (0, "2", "phase1", "comp2"),
                     (0, "2", "phase2", "comp1"), (0, "2", "phase2", "comp2")]
    assert hasattr(m0.fs.u.holdup.port, "sum_split_fractions")
    for k in m0.fs.u.holdup.port.sum_split_fractions.keys():
        assert k in [(0, "phase1", "comp1"), (0, "phase1", "comp2"),
                     (0, "phase2", "comp1"), (0, "phase2", "comp2")]


def test_OutletSplitter_split_material_flow(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "flow"
    m0.fs.u.holdup.port._make_split_fraction()
    m0.fs.u.holdup.port._split_material()

    assert hasattr(m0.fs.u.holdup.port, "material_splitting")
    for k in m0.fs.u.holdup.port.material_splitting.keys():
        assert k in [(0, "1", "phase1", "comp1"),
                     (0, "1", "phase1", "comp2"),
                     (0, "1", "phase2", "comp1"),
                     (0, "1", "phase2", "comp2"),
                     (0, "2", "phase1", "comp1"),
                     (0, "2", "phase1", "comp2"),
                     (0, "2", "phase2", "comp1"),
                     (0, "2", "phase2", "comp2")]


def test_OutletSplitter_split_material_phase(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "phase"
    m0.fs.u.holdup.port._make_split_fraction()
    # Add a phase_component_list to the properties to test this
    m0.fs.u.holdup.config.property_package.phase_component_list = {
            "phase1": ["comp1", "comp2"],
            "phase2": ["comp1"]}
    m0.fs.u.holdup.port._split_material()

    assert hasattr(m0.fs.u.holdup.port, "material_splitting")
    for k in m0.fs.u.holdup.port.material_splitting.keys():
        # Should be no comp2 in phase2 due to phase_component_list
        assert k in [(0, "1", "phase1", "comp1"),
                     (0, "1", "phase1", "comp2"),
                     (0, "1", "phase2", "comp1"),
                     (0, "2", "phase1", "comp1"),
                     (0, "2", "phase1", "comp2"),
                     (0, "2", "phase2", "comp1")]


def test_OutletSplitter_split_material_component(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "component"
    m0.fs.u.holdup.port._make_split_fraction()
    m0.fs.u.holdup.port._split_material()

    assert hasattr(m0.fs.u.holdup.port, "material_splitting")
    for k in m0.fs.u.holdup.port.material_splitting.keys():
        assert k in [(0, "1", "phase1", "comp1"),
                     (0, "1", "phase1", "comp2"),
                     (0, "1", "phase2", "comp1"),
                     (0, "1", "phase2", "comp2"),
                     (0, "2", "phase1", "comp1"),
                     (0, "2", "phase1", "comp2"),
                     (0, "2", "phase2", "comp1"),
                     (0, "2", "phase2", "comp2")]


def test_OutletSplitter_split_material_total(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "total"
    m0.fs.u.holdup.port._make_split_fraction()
    m0.fs.u.holdup.port._split_material()

    assert hasattr(m0.fs.u.holdup.port, "material_splitting")
    for k in m0.fs.u.holdup.port.material_splitting.keys():
        assert k in [(0, "1", "phase1", "comp1"),
                     (0, "1", "phase1", "comp2"),
                     (0, "1", "phase2", "comp1"),
                     (0, "1", "phase2", "comp2"),
                     (0, "2", "phase1", "comp1"),
                     (0, "2", "phase1", "comp2"),
                     (0, "2", "phase2", "comp1"),
                     (0, "2", "phase2", "comp2")]


def test_OutletSplitter_split_energy_phase(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "phase"
    m0.fs.u.holdup.port._make_split_fraction()
    m0.fs.u.holdup.port._split_energy()

    assert hasattr(m0.fs.u.holdup.port, "energy_splitting")
    for k in m0.fs.u.holdup.port.energy_splitting.keys():
        assert k in [(0, "1"),
                     (0, "2")]


def test_OutletSplitter_split_energy_other(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port.config.split_type = "flow"
    m0.fs.u.holdup.port._make_split_fraction()
    m0.fs.u.holdup.port._split_energy()

    assert hasattr(m0.fs.u.holdup.port, "temperature_splitting")
    for k in m0.fs.u.holdup.port.temperature_splitting.keys():
        assert k in [(0, "1"),
                     (0, "2")]


def test_OutletSplitter_split_pressure(m0):
    m0.fs.u.holdup.port = OutletSplitterFrame()
    m0.fs.u.holdup.port.config.outlets = ["1", "2"]
    m0.fs.u.holdup.port._make_references()
    m0.fs.u.holdup.port._build_property_blocks()
    m0.fs.u.holdup.port._split_pressure()

    assert hasattr(m0.fs.u.holdup.port, "pressure_splitting")


def test_OutletSplitter_build(m0):
    m0.fs.u.holdup.port = OutletSplitter(outlets=["1", "2"], split_type="flow")

    assert hasattr(m0.fs.u.holdup.port, "split_fraction")
    for k in m0.fs.u.holdup.port.split_fraction.keys():
        assert k in [(0, "1"), (0, "2")]
    assert hasattr(m0.fs.u.holdup.port, "sum_split_fractions")
    for k in m0.fs.u.holdup.port.sum_split_fractions.keys():
        assert k in [0]
    assert hasattr(m0.fs.u.holdup.port, "material_splitting")
    for k in m0.fs.u.holdup.port.material_splitting.keys():
        assert k in [(0, "1", "phase1", "comp1"),
                     (0, "1", "phase1", "comp2"),
                     (0, "1", "phase2", "comp1"),
                     (0, "1", "phase2", "comp2"),
                     (0, "2", "phase1", "comp1"),
                     (0, "2", "phase1", "comp2"),
                     (0, "2", "phase2", "comp1"),
                     (0, "2", "phase2", "comp2")]
    assert hasattr(m0.fs.u.holdup.port, "temperature_splitting")
    for k in m0.fs.u.holdup.port.temperature_splitting.keys():
        assert k in [(0, "1"),
                     (0, "2")]
    assert hasattr(m0.fs.u.holdup.port, "pressure_splitting")


def test_OutletSplitter_build_no_outlets_arg(m0):
    with pytest.raises(ValueError):
        m0.fs.u.holdup.port = OutletSplitter()
