##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for flowsheet_model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, Constraint, Var
from idaes.core import declare_process_block_class, \
                        PropertyParameterBase, PropertyBlockDataBase

# -----------------------------------------------------------------------------
# Test ParameterBlock
@declare_process_block_class("ParameterBlock")
class _ParameterBlock(PropertyParameterBase):
    def build(self):
        pass

    def get_supported_properties(self):
        return {'simple': {'method': '_simple', 'units': 'mol/s'},
                'recursion1': {'method': '_recursion1', 'units': 'Pa'},
                'recursion2': {'method': '_recursion2', 'units': 'K'},
                'raise_exception': {'method': '_raise_exception',
                                    'units': 'K'},
                'does_not_create_component': {
                        'method': '_does_not_create_component',
                        'units': 'K'}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


def test_PropertyParameterBase():
    # Test that PropertyParameterBase gets module information
    m = ConcreteModel()
    m.p = ParameterBlock()
    super(_ParameterBlock, m.p).build()

    assert hasattr(m.p, "property_module")


# -----------------------------------------------------------------------------
# Test PropertyBlockDataBase
@declare_process_block_class("PropertyBlockData")
class _PropertyBlockData(PropertyBlockDataBase):
    def build(self):
        self._not_callable = 1

    def _simple(self):
        self.simple = Var(initialize=1)

    def _recursion1(self):
        self.recursive_cons1 = Constraint(expr=self.recursion2 == 1)

    def _recursion2(self):
        self.recursive_cons2 = Constraint(expr=self.recursion1 == 1)

    def _raise_exception(self):
        raise KeyError()

    def _does_not_create_component(self):
        pass


def test_PropertyBlock_config():
    # Test that PropertyBlockDataBase config has correct arguments
    m = ConcreteModel()
    m.p = PropertyBlockData()

    assert hasattr(m.p.config, "calculate_equilibrium_reactions")
    assert hasattr(m.p.config, "calculate_phase_equilibrium")
    assert hasattr(m.p.config, "has_sum_fractions")
    assert hasattr(m.p.config, "parameters")

    m.p.config.calculate_equilibrium_reactions = True
    m.p.config.calculate_equilibrium_reactions = False
    with pytest.raises(ValueError):
        m.p.config.calculate_equilibrium_reactions = 'foo'
    with pytest.raises(ValueError):
        m.p.config.calculate_equilibrium_reactions = 10

    m.p.config.calculate_phase_equilibrium = True
    m.p.config.calculate_phase_equilibrium = False
    with pytest.raises(ValueError):
        m.p.config.calculate_phase_equilibrium = 'foo'
    with pytest.raises(ValueError):
        m.p.config.calculate_phase_equilibrium = 10

    m.p.config.has_sum_fractions = True
    m.p.config.has_sum_fractions = False
    with pytest.raises(ValueError):
        m.p.config.has_sum_fractions = 'foo'
    with pytest.raises(ValueError):
        m.p.config.has_sum_fractions = 10


def test_getattr_simple():
    m = ConcreteModel()
    m.pb = ParameterBlock()
    super(_ParameterBlock, m.pb).build()
    m.p = PropertyBlockData(parameters=m.pb)

    m.p.simple_cons = Constraint(expr=m.p.simple == 2)
    assert isinstance(m.p.simple, Var)


def test_getattr_protected():
    m = ConcreteModel()
    m.p = PropertyBlockData()

    with pytest.raises(AttributeError):
        # Call a protected component that does not exist
        m.p.simple_cons = Constraint(expr=m.p._foo == 2)


def test_getattr_recursion():
    m = ConcreteModel()
    m.p = PropertyBlockData()

    with pytest.raises(Exception):
        # Call a component that triggers a recursive loop of calls
        m.p.simple_cons = Constraint(expr=m.p.recursion1 == 2)


def test_getattr_does_not_exist():
    m = ConcreteModel()
    m.p = PropertyBlockData()
    with pytest.raises(AttributeError):
        m.p.simple_cons = Constraint(expr=m.p.does_not_exist == 2)


def test_getattr_not_callable():
    m = ConcreteModel()
    m.p = PropertyBlockData()
    with pytest.raises(AttributeError):
        m.p.simple_cons = Constraint(expr=m.p.not_callable == 2)


def test_getattr_raise_exception():
    m = ConcreteModel()
    m.pb = ParameterBlock()
    super(_ParameterBlock, m.pb).build()
    m.p = PropertyBlockData(parameters=m.pb)
    with pytest.raises(KeyError):
        m.p.simple_cons = Constraint(expr=m.p.raise_exception == 2)


def test_getattr_does_not_create_component():
    m = ConcreteModel()
    m.p = PropertyBlockData()
    with pytest.raises(Exception):
        m.p.simple_cons = Constraint(expr=m.p.does_not_create_component == 2)
