##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""Utility functions for performing partial surrogate.

Implementation of PSC formulation given in...
[citation needed, maybe Quesada and Grossmann 1992]

"""
from __future__ import division, print_function

from pyomo.core.expr.current import SumExpression, identify_variables, \
    clone_expression
from pyomo.environ import (Constraint, ConstraintList, Objective, minimize,
                           value)
from pyomo.repn.canonical_repn import generate_canonical_repn

from idaes.core.util.expr import is_linear

__author__ = "Qi Chen <qichen@andrew.cmu.edu>"


def add_partial_surrogate_cut(model):
    """Apply partial surrogate scheme.

    The partial surrogate approach is in between use of Generalized Benders and
    Outer Approximation. It does not project out as many variables as in GBD,
    but also does not add as many new constraints at each iteration as OA.

    Args:
        model (ConcreteModel): Pyomo model on which to add partial surrogate
            cuts.

    """
    # get ConstraintList in which to store partial surrogate cuts
    try:
        psc_constraint = model.partial_surrogate_cuts
    except AttributeError:
        psc_constraint = model.partial_surrogate_cuts = ConstraintList()

    # get the primary objective function
    objs = model.component_objects(  # iterator over all active objectives
        ctype=Objective, active=True, descend_into=True)
    obj = next(objs, None)  # first active objective encountered
    if obj is None:
        raise ValueError('Model has no active objectives. Cannot generate '
                         'partial surrogate cuts.')
    if next(objs, None) is not None:  # if there are more active objectives...
        raise ValueError('Model has multiple active objectives. Cannot '
                         'generate partial surrogate cuts.')

    # TODO: Need to test for objective sense. Assume minimize for now.
    assert(obj.sense == minimize)
    if not is_linear(obj.expr):
        raise ValueError('PSC implementation does not yet support nonlinear '
                         'objective functions.')

    # require duals information
    if not hasattr(model, 'duals'):
        raise ValueError(
            "Model has no information on duals. User needs to define a Suffix "
            "object on the model named 'duals'.")

    # Right now require IPOPT because we don't know how to get bound
    # multipliers for variables from any other solvers.
    #
    # TODO look up how to get bound multipliers for other solvers
    if not hasattr(model, 'ipopt_zL_out'):
        raise ValueError(
            """Model is missing information on variable bound multipliers.

             User needs to define a Suffix object on the model named
             'ipopt_zL_out'.

            """)
    if not hasattr(model, 'ipopt_zU_out'):
        raise ValueError(
            """Model is missing information on variable bound multipliers.

             User needs to define a Suffix object on the model named
             'ipopt_zU_out'.

            """)

    # nonlinear variables
    nl_vars = nonlinear_vars(model)
    nl_var_ids = set(id(v) for v in nl_vars)

    # generate the sum of all multipliers with the nonlinear constraints
    nl_cons = [c for c in model.component_data_objects(
        ctype=Constraint, active=True, descend_into=True)
        if c.body.polynomial_degree not in (0, 1)]
    # map to substitute the values in for all nonlinear variables
    var_to_val = {id(var): value(var) for var in nl_vars}
    sum_nonlinear = sum(model.duals[c] *
                        (clone_expression(c.body, substitute=var_to_val))
                        for c in nl_cons)
    # TODO handle the case where constraint lower or upper is not 0 or None

    # Generate the sum of all multipliers with linear constraints containing
    # nonlinear variables
    #
    # For now, need to generate canonical representation in order to get the
    # coefficients on the linear terms.
    lin_cons = [c for c in model.component_data_objects(
        ctype=Constraint, active=True, descend_into=True)
        if c.body.polynomial_degree in (0, 1)]
    for constr in lin_cons:
        repn = generate_canonical_repn(constr)
        # create a coefficient dictionary mapping variables to their
        # coefficient in the expression
        coef_dict = {id(var): coef for var, coef in
                     zip(repn.variables, repn.linear)}
    sum_linear = sum(model.duals[c] *
                     sum(coef_dict[id(var)] * (var - value(var))
                         for var in repn.variables
                         if id(var) in nl_var_ids)
                     for c in lin_cons)

    # Generate the sum of all bound multipliers with nonlinear variables
    sum_var_bounds = sum(model.ipopt_zL_out * -1 * (var - value(var)) +
                         model.ipopt_zU_out * (var - value(var))
                         for var in nl_vars)
    # Add everything together
    psc_constraint.add(expr=obj.expr + sum_nonlinear + sum_linear +
                       sum_var_bounds)


def nonlinear_vars(blk):
    """Return the variables that participate in nonlinear terms.

    Args:
        blk (Block): Pyomo block

    Returns:
        list(_VarData): list of Pyomo var_data objects

    """
    nl_vars = []
    # This is a workaround because Var is not hashable, and I do not want
    # duplicates in nl_vars.
    seen = set()
    nl_cons = [c for c in blk.component_data_objects(
        ctype=Constraint, active=True, descend_into=True)
        if c.body.polynomial_degree in (0, 1)]
    for con_data in nl_cons:
        if not con_data.active:
            continue  # if constraint is inactive, skip it
        if isinstance(con_data.body, SumExpression):
            # go through each term and check to see if the term is nonlinear
            for expr in con_data.body.args:
                # Check to see if the expression is nonlinear
                if not is_linear(expr):
                    # collect variables
                    for var in identify_variables(expr, include_fixed=False):
                        if id(var) not in seen:
                            seen.add(id(var))
                            nl_vars.append(var)
        # if the root expression object is not a summation, then something else
        # is the cause of the nonlinearity. Collect all participating
        # variables.
        else:
            # collect variables
            for var in identify_variables(con_data.body, include_fixed=False):
                if id(var) not in seen:
                    seen.add(id(var))
                    nl_vars.append(var)

    return nl_vars
