##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for misc util methods.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import Block, ConcreteModel,  Constraint, \
                            Param, Set, SolverFactory, Var, value
from pyomo.network import Port
from idaes.core.util.misc import add_object_ref, smooth_abs, smooth_minmax, \
                                    smooth_min, smooth_max, \
                                    solve_indexed_blocks, \
                                    fix_port, unfix_port


# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.fixture(scope="module")
def simple_model():
    """Build a simple model for testing.
    """
    m = ConcreteModel()
    m.a = Var(initialize=4.0)
    m.b = Var(initialize=-4.0)
    m.e = Param(default=1e-4)

    return m


def test_add_object_ref_functionality():
    # Create Pyomo objects for testing
    m = ConcreteModel()
    n = ConcreteModel()
    m.a = Var(initialize=4.0)

    # Add reference and assert value
    add_object_ref(n, "a", m.a)

    assert value(n.a) == 4.0


def test_add_object_ref_exception():
    # Test that adding a reference with same name raises AttributeError
    m = ConcreteModel()
    n = ConcreteModel()
    m.a = Var(initialize=4.0)
    n.a = Var(initialize=4.0)

    # Try to create a reference to m.a in a with name a
    with pytest.raises(AttributeError):
        add_object_ref(n, "a", m.a)


def test_smooth_abs_maths():
    # Test basic smooth_abs functionalliy
    assert smooth_abs(4, 0) == 4.0
    assert smooth_abs(-4, 0) == 4.0
    assert smooth_abs(10.0, 0.0) == 10.0
    assert smooth_abs(-10.0, 0.0) == 10.0

    assert smooth_abs(2, 1e-4) == pytest.approx(2.0, abs=1e-4)
    assert smooth_abs(-2, 1e-4) == pytest.approx(2.0, abs=1e-4)
    assert smooth_abs(10) == pytest.approx(10.0, abs=1e-4)
    assert smooth_abs(-10) == pytest.approx(10.0, abs=1e-4)


def test_smooth_abs_expr(simple_model):
    # Test that smooth_abs works with Pyomo components
    assert value(smooth_abs(simple_model.a, 0)) == 4.0
    assert value(smooth_abs(simple_model.b, 0)) == 4.0

    assert (value(smooth_abs(simple_model.a, simple_model.e)) ==
            pytest.approx(4.0, abs=1e-4))
    assert (value(smooth_abs(simple_model.b, simple_model.e)) ==
            pytest.approx(4.0, abs=1e-4))


def test_smooth_abs_a_errors():
    # Test that smooth_abs returns meaningful errors when given invalid arg
    with pytest.raises(TypeError):
        smooth_abs("foo")
    with pytest.raises(TypeError):
        smooth_abs([1, 2, 3])


def test_smooth_abs_eps_errors():
    # Test that smooth_abs returns meaningful errors when given invalid eps
    with pytest.raises(TypeError):
        smooth_abs(1.0, "a")
    with pytest.raises(TypeError):
        smooth_abs(1.0, [1, 2, 3])


def test_smooth_minmax_maths():
    # Test basic smooth_minmax functionality
    assert smooth_minmax(1, 2, 0, sense='max') == 2
    assert smooth_minmax(1, 2, 0, sense='min') == 1
    assert smooth_minmax(5.0, 3, 0.0, sense='max') == 5
    assert smooth_minmax(5.0, 3, 0.0, sense='min') == 3

    assert (smooth_minmax(2.0, 12.0, 1e-4, 'max') ==
            pytest.approx(12.0, abs=1e-4))
    assert (smooth_minmax(2.0, 12.0, 1e-4, 'min') ==
            pytest.approx(2.0, abs=1e-4))
    assert (smooth_minmax(32.0, 12.0, sense='max') ==
            pytest.approx(32.0, abs=1e-4))
    assert (smooth_minmax(32.0, 12.0, sense='min') ==
            pytest.approx(12.0, abs=1e-4))


def test_smooth_minmax_default_sense():
    # Test that smooth_minmax defaults to maximise
    assert smooth_minmax(1, 2, 0,) == 2


def test_smooth_minmax_expr(simple_model):
    # Test that smooth_minmax works with Pyomo components
    assert value(smooth_minmax(simple_model.a,
                               simple_model.b,
                               0,
                               sense='max')) == 4.0
    assert value(smooth_minmax(simple_model.a,
                               simple_model.b,
                               0,
                               sense='min')) == -4.0

    assert value(smooth_minmax(simple_model.a,
                               simple_model.b,
                               sense='max')) == pytest.approx(4.0, abs=1e-4)
    assert value(smooth_minmax(simple_model.a,
                               simple_model.b,
                               sense='min')) == pytest.approx(-4.0, abs=1e-4)

    assert value(smooth_minmax(simple_model.a,
                               simple_model.b,
                               simple_model.e,
                               sense='max')) == pytest.approx(4.0, abs=1e-4)
    assert value(smooth_minmax(simple_model.a,
                               simple_model.b,
                               simple_model.e,
                               sense='min')) == pytest.approx(-4.0, abs=1e-4)


def test_smooth_abs_ab_errors():
    # Test that smooth_abs returns meaningful errors when given invalid args
    with pytest.raises(TypeError):
        smooth_abs("foo", 1)
    with pytest.raises(TypeError):
        smooth_abs(3, [1, 2, 3])


def test_smooth_minmax_eps_errors():
    # Test that smooth_abs returns meaningful errors when given invalid eps
    with pytest.raises(TypeError):
        smooth_minmax(1.0, 1.0, "foo")
    with pytest.raises(TypeError):
        smooth_minmax(1.0, 1.0, [1, 2, 3])


def test_smooth_minmax_sense_errors():
    # Test that smooth_abs returns meaningful errors when given invalid sense
    with pytest.raises(ValueError):
        smooth_minmax(1.0, 1.0, sense="foo")
    with pytest.raises(ValueError):
        smooth_minmax(1.0, 1.0, sense=1.0)
    with pytest.raises(ValueError):
        smooth_minmax(1.0, 1.0, sense=[1.0])


def test_smooth_max(simple_model):
    # Test that smooth_max gives correct values
    assert smooth_max(3.0, 12.0) == pytest.approx(12.0, abs=1e-4)
    assert value(smooth_max(simple_model.a,
                            simple_model.b,
                            simple_model.e)) == pytest.approx(4.0, abs=1e-4)


def test_smooth_min(simple_model):
    # Test that smooth_min gives correct values
    assert smooth_min(3.0, 12.0) == pytest.approx(3.0, abs=1e-4)
    assert value(smooth_min(simple_model.a,
                            simple_model.b,
                            simple_model.e)) == pytest.approx(-4.0, abs=1e-4)


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_solve_indexed_block_list():
    # Create an indexed block and try to solve it
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3])

    def block_rule(b, x):
        b.v = Var(initialize=1.0)
        b.c = Constraint(expr=b.v == 2.0)
    m.b = Block(m.s, rule=block_rule)

    solve_indexed_blocks(solver=solver, blocks=[m.b])

    for i in m.s:
        assert value(m.b[i].v == 2.0)


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_solve_indexed_block_IndexedBlock():
    # Create an indexed block and try to solve it
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3])

    def block_rule(b, x):
        b.v = Var(initialize=1.0)
        b.c = Constraint(expr=b.v == 2.0)
    m.b = Block(m.s, rule=block_rule)

    solve_indexed_blocks(solver=solver, blocks=m.b)

    for i in m.s:
        assert value(m.b[i].v == 2.0)


def test_solve_indexed_block_error():
    # Try solve_indexed_block on non-block object
    with pytest.raises(TypeError):
        solve_indexed_blocks(solver=None, blocks=[1, 2, 3])


def test_fix_port_no_args():
    m = ConcreteModel()
    m.v = Var(initialize=1.0)
    m.c = Port(noruleinit=True)
    m.c.add(m.v, "v")

    fix_port(m.c, "v")

    assert m.v.fixed is True
    assert m.v.value == 1.0


def test_fix_port_value():
    m = ConcreteModel()
    m.v = Var(initialize=1.0)
    m.c = Port(noruleinit=True)
    m.c.add(m.v, "v")

    fix_port(m.c, "v", value=2.0)

    assert m.v.fixed is True
    assert m.v.value == 2.0


def test_fix_port_indexed():
    m = ConcreteModel()
    m.v1 = Var(initialize=1.0)
    m.v2 = Var(initialize=10.0)
    m.c = Port([1, 2], noruleinit=True)
    m.c[1].add(m.v1, "v1")
    m.c[2].add(m.v2, "v2")

    fix_port(m.c, "v1", port_idx=[1])

    assert m.v1.fixed is True
    assert m.v2.fixed is False
    assert m.v1.value == 1.0
    assert m.v2.value == 10.0


def test_fix_port_indexed_value():
    m = ConcreteModel()
    m.v1 = Var(initialize=1.0)
    m.v2 = Var(initialize=10.0)
    m.c = Port([1, 2], noruleinit=True)
    m.c[1].add(m.v1, "v1")
    m.c[2].add(m.v2, "v2")

    fix_port(m.c, "v2", port_idx=[2], value=30.0)

    assert m.v1.fixed is False
    assert m.v2.fixed is True
    assert m.v1.value == 1.0
    assert m.v2.value == 30.0


def test_fix_port_indexed_var():
    m = ConcreteModel()
    m.v = Var([1, 2], initialize=1.0)
    m.c = Port(noruleinit=True)
    m.c.add(m.v, "v")

    fix_port(m.c, "v", comp=1)

    assert m.v[1].fixed is True
    assert m.v[2].fixed is False
    assert m.v[1].value == 1.0
    assert m.v[2].value == 1.0


def test_fix_port_indexed_var_value():
    m = ConcreteModel()
    m.v = Var([1, 2], initialize=1.0)
    m.c = Port(noruleinit=True)
    m.c.add(m.v, "v")

    fix_port(m.c, "v", comp=1, value=2.0)

    assert m.v[1].fixed is True
    assert m.v[2].fixed is False
    assert m.v[1].value == 2.0
    assert m.v[2].value == 1.0


def test_fix_port_both_indexed():
    m = ConcreteModel()
    m.v1 = Var([1, 2], initialize=1.0)
    m.v2 = Var(initialize=1.0)
    m.c = Port([1, 2], noruleinit=True)
    m.c[1].add(m.v1, "v1")
    m.c[2].add(m.v2, "v2")

    fix_port(m.c, "v1", port_idx=[1], comp=1)

    assert m.v1[1].fixed is True
    assert m.v1[2].fixed is False
    assert m.v2.fixed is False
    assert m.v1[1].value == 1.0
    assert m.v1[2].value == 1.0
    assert m.v2.value == 1.0


def test_fix_port_both_indexed_value():
    m = ConcreteModel()
    m.v1 = Var([1, 2], initialize=1.0)
    m.v2 = Var(initialize=1.0)
    m.c = Port([1, 2], noruleinit=True)
    m.c[1].add(m.v1, "v1")
    m.c[2].add(m.v2, "v2")

    fix_port(m.c, "v1", port_idx=[1], comp=1, value=10.0)

    assert m.v1[1].fixed is True
    assert m.v1[2].fixed is False
    assert m.v2.fixed is False
    assert m.v1[1].value == 10.0
    assert m.v1[2].value == 1.0
    assert m.v2.value == 1.0


def test_unfix_port_no_args():
    m = ConcreteModel()
    m.v = Var(initialize=1.0)
    m.v.fix()
    m.c = Port(noruleinit=True)
    m.c.add(m.v, "v")

    unfix_port(m.c, "v")

    assert m.v.fixed is False
    assert m.v.value == 1.0


def test_unfix_port_indexed():
    m = ConcreteModel()
    m.v1 = Var(initialize=1.0)
    m.v2 = Var(initialize=10.0)
    m.v1.fix()
    m.v2.fix()
    m.c = Port([1, 2], noruleinit=True)
    m.c[1].add(m.v1, "v1")
    m.c[2].add(m.v2, "v2")

    unfix_port(m.c, "v1", port_idx=[1])

    assert m.v1.fixed is False
    assert m.v2.fixed is True
    assert m.v1.value == 1.0
    assert m.v2.value == 10.0


def test_unfix_port_indexed_var():
    m = ConcreteModel()
    m.v = Var([1, 2], initialize=1.0)
    m.v.fix()
    m.c = Port(noruleinit=True)
    m.c.add(m.v, "v")

    unfix_port(m.c, "v", comp=1)

    assert m.v[1].fixed is False
    assert m.v[2].fixed is True
    assert m.v[1].value == 1.0
    assert m.v[2].value == 1.0


def test_unfix_port_both_indexed():
    m = ConcreteModel()
    m.v1 = Var([1, 2], initialize=1.0)
    m.v2 = Var(initialize=1.0)
    m.v1.fix()
    m.v2.fix()
    m.c = Port([1, 2], noruleinit=True)
    m.c[1].add(m.v1, "v1")
    m.c[2].add(m.v2, "v2")

    unfix_port(m.c, "v1", port_idx=[1], comp=1)

    assert m.v1[1].fixed is False
    assert m.v1[2].fixed is True
    assert m.v2.fixed is True
    assert m.v1[1].value == 1.0
    assert m.v1[2].value == 1.0
    assert m.v2.value == 1.0
