##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Data Management Framework
"""
# stdlib
import logging
import os
import shutil
import sys
# third-party
import pendulum
import six
from traitlets import HasTraits, default, observe
from traitlets import Unicode
import yaml
# local
from . import errors, workspace, resource, resourcedb, propdata
from .util import get_logger

_log = get_logger('dmf')


class DMFConfig(object):
    """Global DMF configuration.

    Every time you create an instance of the :class:`DMF <idaes.dmf.dmf.DMF>`,
    or run a ``dmf`` command on the command-line, the library opens the
    global DMF configuration file to figure out wthe default workspace
    (and, eventually, other values).

    The default location for this configuration file is "~/.dmf", i.e.
    the file named ".dmf" in the user's home directory. This can be
    modified programmatically by changing the "filename" attribute of
    this class.

    The contents of .dmf are formatted as `YAML`_, with the following
    keys defined:

        workspace
            Path to the default workspace directory.

    An example file is shown below:

    .. code-block:: YAML

        {workspace: /tmp/newdir}

    .. _YAML: http://www.yaml.org/
    """
    # configuration file location
    filename = os.path.expanduser('~/.dmf')
    # known keys
    WORKSPACE = 'workspace'
    _keys = [WORKSPACE]
    # defaults
    DEFAULTS = {WORKSPACE: os.path.abspath('.')}

    def __init__(self, defaults=None):
        """Create new configuration.

        Args:
            defaults (dict): Default values to use if nothing is found.
                            If the value is 'True', use self.DEFAULTS
        Raises:
            IOError: If self.filename is not openable or parse-able
        """
        self.c, fp = {}, None
        try:
            fp = open(self.filename, 'rb')
        except IOError:
            if defaults is None:
                raise
            if defaults is True:
                self.c.update(self.DEFAULTS)
            else:
                self.c.update(defaults)
        if fp:
            try:
                self._parse(fp)
            except ValueError as err:
                raise IOError('Parsing configuration at "{}": {}'
                              .format(self.filename, err))

    def save(self):
        try:
            fp = open(self.filename, 'w')
        except IOError:
            raise
        try:
            self._write(fp)
        except ValueError as err:
            raise IOError('Failed to write in YAML format to <{}>: {}'
                          .format(self.filename, err))

    @property
    def workspace(self):
        return self.c[self.WORKSPACE]

    def _parse(self, fp):
        try:
            y = yaml.load(fp)
        except Exception as err:
            raise ValueError(err)
        if y:
            for k, v in six.iteritems(y):
                if k in self._keys:
                    self.c[k] = v

    def _write(self, fp):
        yaml.dump(self.c, fp)


class DMF(workspace.Workspace, HasTraits):
    """Data Management Framework (DMF).

    Expected usage is to instantiate this class, once, and then use it for
    storing, searching, and retrieve :term:`resource` \s that
    are required for the given analysis.

    For details on the configuration files used by the DMF, see
    documentation for :class:`DMFConfig` (global configuration) and
    :class:`idaes.dmf.workspace.Workspace`.
    """
    db_file = Unicode(help='Database file name')
    datafile_dir = Unicode(help='Data file directory, '
                                'relative to DMF root')

    CONF_DB_FILE = 'db_file'
    CONF_DATA_DIR = 'datafile_dir'
    CONF_HELP_PATH = workspace.Fields.DOC_HTML_PATH

    # logging should really provide this
    _levelnames = {'fatal': logging.FATAL, 'error': logging.ERROR,
                   'warn': logging.WARN, 'warning': logging.WARN,
                   'info': logging.INFO, 'debug': logging.DEBUG}

    def __init__(self, path='', name=None, desc=None, **ws_kwargs):
        # type: (str, ...) -> None
        """

        Args:
            path: Path to workspace. If given, will override any
                  global configuration. If not given (empty or None), then
                  global configuration will be checked first. If empty, and
                  the global configuration does not parse or exist, then
                  "." will be used.
            name (str): Name to be used for workspace.
            desc (str): Longer description of workspace.
            **ws_kwargs: Keyword arguments for :meth:`workspace.Workspace()`
                         constructor.
        Raises:
            errors.DMFWorkspaceNotFoundError: If workspace is not found, and
                 the `create` option was not given in the keyword args.
            errors.DMFBadWorkspaceError: If workspace is found, but there
                 are errors with its configuration.
        """
        # get global configuration
        conf = DMFConfig(defaults={})
        # get path, if not specified, from configuration
        if not path:
            path = conf.c.get(DMFConfig.WORKSPACE, '.')
        # set up workspace
        try:
            workspace.Workspace.__init__(self, path, **ws_kwargs)
        except OSError as err:
            raise errors.DMFBadWorkspaceError(path,
                                              'Cannot create: {}'.format(err))
        except errors.WorkspaceNotFoundError:
            raise errors.DMFWorkspaceNotFoundError(path)
        except errors.WorkspaceConfNotFoundError:
            raise errors.DMFBadWorkspaceError(path, 'Configuration not found')
        except errors.WorkspaceConfMissingField as err:
            raise errors.DMFBadWorkspaceError(path, str(err))
        try:
            meta_dict = self.meta  # call associated code
        except (errors.ParseError, ValueError) as err:
            msg = 'Configuration parse error: {}'.format(err)
            raise errors.DMFBadWorkspaceError(path, msg)
        try:
            self._validate_conf(meta_dict)
        except ValueError as err:
            msg = 'Configuration validation error: {}'.format(err)
            raise errors.DMFBadWorkspaceError(path, msg)
        # set up logging
        if workspace.Fields.LOG_CONF in meta_dict:
            try:
                self._configure_logging(meta_dict[workspace.Fields.LOG_CONF])
            except ValueError as err:
                msg = 'Configuration, logging section, error: {}'.format(err)
                raise errors.DMFBadWorkspaceError(path, msg)
        # set up rest of DMF
        path = os.path.join(self.root, self.db_file)
        self._db = resourcedb.ResourceDB(path)
        self._datafile_path = os.path.join(self.root, self.datafile_dir)
        if not os.path.exists(self._datafile_path):
            os.mkdir(self._datafile_path, 0o750)
        # add create/modified date, and optional name/description
        _w = workspace.Workspace
        right_now = pendulum.now().to_datetime_string()
        meta = {_w.CONF_CREATED: right_now, _w.CONF_MODIFIED: right_now}
        if name:
            meta[_w.CONF_NAME] = name
        if desc:
            meta[_w.CONF_DESC] = desc
        self.set_meta(meta)

    def _configure_logging(self, conf):
        """Configure logging for DMF.

        Expected schema::

            <logger-name>:
                level: <levelname>              default=NOTSET
                output: file|_stdout_|_stderr_  default=_stderr_

        Args:
            conf (dict): Configuration dict
        Raises:
            ValueError: for bad configuration values
        """
        for lognm in conf.keys():
            name = lognm.lower()
            if name == 'root':
                log = get_logger()
            else:
                log = get_logger(lognm)
            subconf = conf[lognm]
            h = logging.StreamHandler()
            if 'output' in subconf:
                dest = subconf['output']
                if dest == '_stdout_':
                    h = logging.StreamHandler(stream=sys.stdout)
                elif dest == '_stderr_':
                    h = logging.StreamHandler(stream=sys.stderr)
                else:
                    try:
                        h = logging.FileHandler(dest)
                    except IOError:
                        raise ValueError('Cannot open output file "{}" for '
                                         'logger "{}"'.format(dest, lognm))
                log.addHandler(h)
            log.setLevel(logging.NOTSET)
            if 'level' in subconf:
                levelnm = subconf['level'].lower()
                level = self._levelnames.get(levelnm, None)
                if level is None:
                    opt = ', '.join(self._levelnames.keys())
                    raise ValueError('Bad level "{}" for logger "{}". Must be '
                                     'one of: {}'.format(levelnm, lognm, opt))
                log.setLevel(level)

    def _validate_conf(self, c):
        if self.CONF_HELP_PATH not in c:
            _log.warn('Path to built HTML documentation is not set. '
                      'The DMF "help" command will not work. To set '
                      'this path, set "{}" in the DMF configuration file.'
                      .format(self.CONF_HELP_PATH))

    @default(CONF_DB_FILE)
    def _default_db_file(self):
        return self.meta.get(self.CONF_DB_FILE, 'resourcedb.json')

    @default(CONF_DATA_DIR)
    def _default_res_dir(self):
        return self.meta.get(self.CONF_DATA_DIR, 'files')

    @observe(CONF_DB_FILE, CONF_DATA_DIR, CONF_HELP_PATH)
    def _observe_setting(self, change):
        if change['type'] != 'change':
            return
        values = {change['name']: change['new']}
        self.set_meta(values)

    def add(self, rsrc):
        """Add a resource and associated files.


        Args:
            rsrc (resource.Resource): The resource
        Returns:
            (str) Resource ID
        Raises:
            DMFError, DuplicateResourceError
        """
        # Copy files as necessary
        # Note: this updates paths in the Resource, so should come first
        if rsrc.datafiles:
            self._copy_files(rsrc.datafiles, rsrc.datafiles_dir)
        # Add resource
        try:
            self._db.put(rsrc)
        except errors.DuplicateResourceError as err:
            _log.error('Cannot add resource: {}'.format(err))
            raise
        return rsrc.id_

    def _copy_files(self, files, subdir):
        dest_dir = os.path.join(self._datafile_path, subdir)
        if not os.path.exists(dest_dir):
            try:
                os.mkdir(dest_dir)
            except os.error as err:
                raise errors.DMFError('Cannot make destination dir "{}" '
                                      'for datafiles: {}'
                                      .format(dest_dir, err))
        for f in filter(lambda _: _.do_copy, files):
            src = f.fullpath
            # skip, if already in the dest directory
            if os.path.dirname(src) == dest_dir:
                continue
            dest_name = os.path.basename(src)
            # copy to new directory
            dest = os.path.join(dest_dir, dest_name)
            _log.debug('Copy "{}" => "{}"'.format(src, dest))
            shutil.copy2(src, dest)
            # update resource
            f.path = dest_name
            f.root = self._datafile_path
            f.subdir = subdir
            _log.debug('Resource updated: root={} name={} full-path="{}"'
                       .format(f.root, f.path, f.fullpath))
            # remove original, if temporary
            if f.is_tmp:
                try:
                    os.unlink(src)
                except OSError as err:
                    _log.error('Unable to remove temporary datafile "{}": {}'
                               .format(src, err))
                    # note: not a fatal error

    def count(self):
        return len(self._db)

    def fetch_one(self, rid):
        """Fetch one resource, from its identifier.

        Args:
            rid (int): Resource identifier
        Returns:
            (resource.Resource) The found resource, or None if no match
        """
        return self._postproc_resource(self._db.get(rid))

    def fetch_many(self, rid_list):
        """Fetch multiple resources, by their identifiers.

        Args:
            rid_list (list): List of integer resource identifers

        Returns:
            (list of resource.Resource) List of found resources (may be empty)
        """
        for rid in rid_list:
            yield self.fetch_one(rid)

    def find(self, filter_dict=None, id_only=False):
        """Find and return resources matching the filter.

        The filter syntax is a subset of the MongoDB filter syntax.
        This means that it is represented as a dictionary, where
        each key is an attribute or nested attribute name, and each
        value is the value against which to match. There are four
        possible types of values:

        1. scalar string or number (int, float): Match resources that
           have this exact value for the given attribute.
        2. date, as datetime.datetime or pendulum.Pendulum instance: Match
           resources that have this exact date for the given attribute.
        3. list: Match resources that have a list value for this attribute,
           and for which any of the values in the provided list are in the
           resource's corresponding value. If a '!' is appended to the key
           name, then this will be interpreted as a directive to only match
           resources for which *all* values in the provided list are present.
        4. dict: This is an inequality, with one or more key/value pairs.
           The key is the type of inequality and the value is the numeric
           value for that range. All keys begin with '$'. The possible
           inequalities are:

                - "$lt": Less than (<)
                - "$le": Less than or equal (<=)
                - "$gt": Greater than (>)
                - "$ge": Greater than or equal (>=)
                - "$ne": Not equal to (!=)

        Args:
            filter_dict (dict): Search filter.
            id_only (bool): If true, return only the identifier of each
                resource; otherwise a Resource object is returned.

        Returns:
            (list of int|Resource) Depending on the value of `id_only`.
        """
        return (self._postproc_resource(r) for r in
                self._db.find(filter_dict=filter_dict, id_only=id_only))

    def find_related(self, rsrc, filter_dict=None, maxdepth=0, meta=None,
                     outgoing=True):
        """Find related resources.

        Args:
            rsrc (resource.Resource): Resource starting point
            filter_dict (dict): See parameter of same name in :meth:`find`.
            maxdepth (int): Maximum depth of search (starts at 1)
            meta (List[str]): Metadata fields to extract for meta part
            outgoing (bool): If True, look at outgoing relations. Otherwise
                             look at incoming relations. e.g. if A 'uses' B
                             and if True, would find B starting from A.
                             If False, would find A starting from B.
        Returns:
            Generates  triples (depth, Triple, meta), where
            the depth is an integer (starting at 1), the Triple is a
            simple namedtuple wrapping (subject, object, predicate), and
            `meta` is a dict of metadata for the endpoint of the relation
            (the object if outgoing=True, the subject if outgoing=False)
            for the fields provided in the `meta` parameter.
        Raises:
            NoSuchResourceError: if the starting resource is not found
        """
        try:
            return self._db.find_related(rsrc.uuid, outgoing=outgoing,
                                         maxdepth=maxdepth, meta=meta)
        except KeyError:
            raise errors.NoSuchResourceError(id_=rsrc.id_)

    def remove(self, identifier=None, filter_dict=None, update_relations=True):
        """Remove one or more resources, from its identifier or a filter.
        Unless told otherwise, this method will scan the DB and remove
        all relations that involve this resource.

        Args:
            identifier (int): Identifier in `id_` attribute
              of a resource.
            filter_dict (dict): Filter to use instead of identifier
            update_relations (bool): If True (the default), scan the DB and
                remove all relations that involve this identifier.
        """
        if not any((identifier, filter_dict)):
            return None
        if identifier is not None:
            try:
                rsrc = self._db.get(identifier)
            except errors.NoSuchResourceError:
                _log.error('Cannot find resource id={} to remove'.format(
                    identifier))
                return None
            if update_relations:
                for rsrc2 in self.find():
                    keep = [rel for rel in rsrc2.relations
                            if rel.object != rsrc.uuid]
                    if len(keep) < len(rsrc2.relations):
                        rsrc2.relations = keep
                        self.update(rsrc2)
            self._db.delete(identifier)
        else:
            idlist = self.find(filter_dict=filter_dict, id_only=True)
            for eid in idlist:
                self.remove(identifier=eid, update_relations=update_relations)

    def update(self, rsrc, sync_relations=False, upsert=False):
        """Update/insert stored resource.

        Args:
            rsrc (resource.Resource): Resource instance
            sync_relations (bool): If True, and if resource
                exists in the DB, then the "relations" attribute of
                the provided resource will be changed to the stored value.
            upsert (bool): If true, and the resource is not in the DMF,
                           then insert it. If false, and the resource is
                           not in the DMF, then do nothing.
        Returns:
            bool: True if the resource was updated or added, False if nothing
                  was done.
        Raises:
            errors.DMFError: If the input resource was invalid.
        """
        did_update = False
        # sanity-check input
        if not isinstance(rsrc, resource.Resource):
            raise TypeError('Resource type expected, got: {}'.format(
                type(rsrc)))
        # synchronize relations
        if sync_relations:
            db_rsrc = self._db.get(rsrc.id_)
            if db_rsrc is not None:
                rsrc.relations = db_rsrc.relations
        # update or insert new values
        try:
            exists = self._db.update(rsrc.id_, rsrc.as_dict())
            if exists is None:
                if upsert:
                    self._db.put(rsrc)
                    did_update = True
            else:
                did_update = True
        except ValueError as err:
            raise errors.DMFError('Bad value for new resource: {}'
                                  .format(err))
        return did_update

    def _postproc_resource(self, r):
        """Perform any additional changes to resources retrieved,
        e.g. to match the configuration in this DMF, before passing
        them up to the application.
        """
        if r and not isinstance(r, int):
            # update relative paths
            for p in r.datafiles:
                p.root = self._datafile_path
        return r

    def __str__(self):
        return 'DMF config="{}"'.format(self._conf)


def get_propertydb_table(rsrc):
    return propdata.PropertyTable.load(rsrc.datafiles[0].fullpath)

