##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for idaes.dmf.resource module
"""
# stdlib
import datetime
import json
import os
import tempfile
import time
#
import pendulum
import pytest
import six
import tinydb
from traitlets import HasTraits, TraitError, List, Unicode, Dict
#
from idaes.dmf import dmf, resource, resourcedb, propdata, util

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'
__version__ = '0.0.1'


@pytest.fixture
def example_resource():
    return _make_example_resource()

g_ctr = 1


def _make_example_resource():
    r, ver = _make_basic_resource('code')
    r.collaborators = [resource.Contact(name='Clark Kent',
                                        email='ckent@dailyplanet.org'),
                       resource.Contact(name='Superman',
                                        email='sman@fortress.solitude.edu')]
    r.sources = [resource.Source(isbn='978-0201616224',
                                 source='Hunt, A. and Thomas, D., '
                                        '"The Pragmatic Programmer", '
                                        'Addison-Wesley, 1999',
                                 date='1999-01-01')]
    r.codes = [resource.Code(type='function', name='test_resource_full',
                             description='The test function',
                             release=ver, location=__file__)]
    r.datafiles = [resource.FilePath(path='/etc/passwd')]
    r.aliases = ['test_resource_full']
    r.tags = ['test', 'resource']
    # r.relations  -- deal with this separately
    r.data = {
        'arbitrary': {
            'values': [1, 2, 3]
        },
    }
    return r


def _make_basic_resource(typename):
    ver = resource.Version(revision=__version__)
    r = resource.Resource(typename, id_=g_ctr)
    _update_counter()
    r.version = ver
    r.creator = resource.Contact(name=__author__.split('<')[0],
                                 email=__author__.split('<')[1][:-1])
    return r, ver


def _update_counter():
    global g_ctr
    g_ctr += 1


@pytest.fixture(scope='function')
def temp_tinydb():
    fp, name = tempfile.mkstemp()
    db = tinydb.TinyDB(name)
    yield db
    # teardown
    db.close()
    os.unlink(name)


####


def test_resource_full(example_resource):
    """Build up a resource with all attributes,
    then make sure it serializes and deserializes.
    """
    r = example_resource  # alias
    # now: let's play
    r_d = r.as_dict()
    # make sure we can serialize it
    j_d = json.dumps(r_d, indent=2)
    #print('@@ serialized to:\n{}\n'.format(j_d))
    r2 = resource.Resource.from_dict(r_d)
    assert r2.version == r.version
    assert r2.sources[0].date == r.sources[0].date
    assert r2.data['arbitrary']['values'][0] == 1
    #
    r2_d = r2.as_dict()
    r2_d2 = resource.Resource.from_dict(r2_d)
    assert r2_d2.version == r.version
    assert r2_d2.sources[0].date == r.sources[0].date
    assert r2_d2.data['arbitrary']['values'][0] == 1


def test_resourcedb_insert(example_resource, temp_tinydb):
    r, tdb = example_resource, temp_tinydb  # aliases
    db = resourcedb.ResourceDB(None, connection=tdb)
    db.put(r)


def test_resourcedb_insert_perf(example_resource, temp_tinydb):
    r, tdb = example_resource, temp_tinydb  # aliases
    db = resourcedb.ResourceDB(None, connection=tdb)
    nn, n, ni = 200, 50, 50
    while n <= nn:
        i = 0
        t0 = pendulum.now()
        while i < ni:
            db.put(r)
            i += 1
        t1 = pendulum.now()
        sec = (t1 - t0).as_timedelta().total_seconds()
        rate = ni / sec
        dur = sec / ni
        assert rate > 1.0
        n1, n2 = n - ni, n
        print('{:d} - {:d}: {:.6f} sec/insert = {:.0f} inserts/sec'
              .format(n1, n2, dur, rate))
        n += ni  # ni! ni! ni!


def test_resourcedb_get(example_resource, temp_tinydb):
    r, tdb = example_resource, temp_tinydb  # aliases
    db = resourcedb.ResourceDB(None, connection=tdb)
    db.put(r)
    r_id = r.id_
    #print('@@ get resource id={}'.format(r_id))
    r2 = db.get(r_id)
    #print('@@ got resource id={}'.format(r2.id_))
    assert r2.id_ == r_id
    assert r2.data['arbitrary']['values'][0] == 1


def test_resourcedb_search(example_resource, temp_tinydb):
    # setup
    r, tdb = example_resource, temp_tinydb  # aliases
    # insert 5 records
    db = resourcedb.ResourceDB(None, connection=tdb)
    _insert_n_records(db, r, 5)
    assert len(db) == 5
    # search for records
    all_ = db.find({})
    assert len(list(all_)) == 5
    some = db.find({'tags': ['iteration1', 'iteration2']})
    assert len(list(some)) == 2
    nope = db.find({'tags!': ['iteration1', 'iteration2']})
    assert len(list(nope)) == 0
    some = db.find({'tags!': ['iteration1', 'test', 'resource']})
    assert len(list(some)) == 1


def test_resourcedb_delete(example_resource, temp_tinydb):
    # setup
    r, tdb = example_resource, temp_tinydb  # aliases
    # insert 5 records
    db = resourcedb.ResourceDB(None, connection=tdb)
    _insert_n_records(db, r, 5)
    assert len(db) == 5
    # find one of them
    r3 = six.next(db.find({'tags': ['iteration3']}))
    id_ = r3.id_
    # delete it
    db.delete(id_)
    # try again to find it
    try:
        six.next(db.find({'tags': ['iteration3']}))
        assert False
    except StopIteration:
        pass


def _insert_n_records(db, r, n=5):
    orig_tags = list(r.tags)
    for i in range(n):
        r.tags = ['iteration{:d}'.format(i)] + orig_tags
        db.put(r)




_propd = [
        {
            "name": "Viscosity Value",
            "units": "mPa-s",
            "values": [2.6, 6.2],
            "error_type": "absolute",
            "errors": [0.06, 0.004],
            "type": "property"
        },
        {
            "name": "r",
            "units": "",
            "values": [0.2, 1000],
            "type": "state"
        }
    ]

_propm = {
    "datatype": "MEA",
    "info": "J. Chem. Eng. Data, 2009, Vol 54, pg. 3096-30100",
    "notes": "r is MEA weight fraction in aqueous soln.",
    "authors": "Amundsen, T.G., Lars, E.O., Eimer, D.A.",
    "title": "Density and Viscosity of Monoethanolamine + .etc.",
    "date": "1970-01-01"
}


def test_property_data_resource():
    workdir = tempfile.mkdtemp()
    dmf_ = dmf.DMF(workdir, create=True)

    pm = propdata.PropertyMetadata(_propm)
    pd = propdata.PropertyData(_propd)
    tbl = propdata.PropertyTable(data=pd, metadata=pm)
    rsrc = resource.PropertyDataResource(property_table=tbl)

    #print('@@ table1 properties = {}'.format(tbl.data.properties))

    my_tag = 'test_propdata'
    rsrc.tags.append(my_tag)
    dmf_.add(rsrc)

    rlist = list(dmf_.find({'tags': [my_tag]}))
    assert len(rlist) == 1
    r1 = rlist[0]
    #print('@@ DUMP of Resource:\n{}'.format(r1))
    tbl2 = r1.property_table
    #print('@@ table2 properties = {}'.format(tbl2.data.properties))

    assert tbl.data.properties[0]['values'] == \
           tbl2.data.properties[0]['values']
    assert tbl.data.properties[0]['errors'] == \
           tbl2.data.properties[0]['errors']
    assert tbl.metadata[0].title == tbl2.metadata[0].title


def test_populate_resource():
    # Create an empty resource
    rsrc = resource.Resource(type='plot')

    # Populate the resource. All the attributes are optional.
    # This is set automatically to current time
    rsrc.created = '2017-10-31'
    # This is set automatically to current time
    rsrc.modified = '2017-11-23'
    # Description of the resource
    rsrc.desc = 'Plots of the Iris dataset'
    # Version of the resource
    rsrc.version = resource.Version(revision='1.0.3-a7',
                                    name='november-release')
    # Name and contact email of creator of the resource
    rsrc.creator = resource.Contact(name='Dan Gunter',
                                    email='dang@science-lab.gov')
    # Name and contact email of other people involved
    rsrc.collaborators = [resource.Contact(name='John Eslick',
                                           email='john@science-lab.gov'),
                          resource.Contact(name='David Miller',
                                           email='david@science-lab.gov')]
    # Provenance: sources, such as publications -- but
    # really anything is allowed -- for the resource.
    rsrc.sources = [resource.Source(
        source='R. A. Fisher. "The use of multiple measurements '
               'in taxonomic problems". '
               'Annals of Eugenics. 7 (2): 179-188.',
        doi='10.1111/j.1469-1809.1936.tb02137.x', date='1936-01-01'),
        resource.Source(source='Edgar Anderson. '
                               '"The species problem in Iris". '
                               'Annals of the Missouri Botanical Garden. '
                               '23 (3): 457-509. '
                               'JSTOR 2394164',
                        date='1936-01-01')]
    # Associated code files, including Jupyter notebooks.
    # If these codes are worth making into their own Resources,
    # you should not put them here, but instead link to them
    # through the `relations` attribute.
    rsrc.codes = [resource.Code(type='notebook',
                                desc='Python plotting code',
                                name='plot_iris_dataset.ipynb',
                                language='Python',
                                location='http://scikit-learn.org/stable/_downloads/'
                                         'plot_iris_dataset.ipynb')]
    # Associated data files. Same comment as for `codes` re: linking.
    rsrc.datafiles = [resource.FilePath(path='iris-data.csv',
                                        desc='Iris data')]
    # Short names that make it easier to find this resource.
    rsrc.aliases = ['iris', 'iris-plots']
    # Tags that make it easier to find this resource later.
    rsrc.tags = ['iris', 'anderson', 'fisher']
    # Arbitrary additional data to add into the resource.
    # Note that larger data units should be made into files and
    # added into the `datafiles` attributes
    rsrc.data = {
        'features': ['sepal length (cm)', 'sepal width (cm)',
                     'petal length (cm)', 'petal width (cm)'],
        # the data is already defined in the datafile,
        # but we could also put it here, inline
        'array': [[5.1, 3.5, 1.4, 0.2], [4.9, 3.0, 1.4, 0.2],
                  [4.7, 3.2, 1.3, 0.2], [4.6, 3.1, 1.5, 0.2],
                  [5.0, 3.6, 1.4, 0.2], [5.4, 3.9, 1.7, 0.4],
                  # ... you get the idea..
                  ]
    }
    # Check the serialization methods
    s = str(rsrc)
    assert 'Annals of the Missouri Botanical Garden' in s
    s = rsrc._repr_html_()
    assert 'Annals of the Missouri Botanical Garden' in s


def test_resource_data_mimetype(example_resource):
    r = example_resource
    r.data = {'table1': 'Name,Age\n"Kid1",14\n"Kid2",12\n'}
    s = r.as_dict()
    # CSV is in serialized data
    assert r.data['table1'] == s['data']['table1']


def test_datetime_trait_set():
    import datetime

    class Dummy(HasTraits):
        date = resource.DateTime()

    t = datetime.datetime.now()
    tp = (t.year, t.month, t.day, t.hour, t.minute, t.second, t.microsecond)
    dumb = Dummy()
    target = pytest.approx(util.datetime_timestamp(t))
    # set by tuple
    dumb.date = tp
    assert dumb.date == target
    with pytest.raises(TraitError):
        dumb.date = (1e15, 0, 0)
    with pytest.raises(TraitError):
        dumb.date = ('not', 'a', 'date')
    # set by str
    dumb.date = t.isoformat()
    assert dumb.date == target
    with pytest.raises(TraitError):
        dumb.date = '1972 was a long time ago'
    # set by timestamp
    dumb.date = util.datetime_timestamp(t)
    assert dumb.date == target
    # set some illegal values
    with pytest.raises(TraitError):
        dumb.date = 1e15
    with pytest.raises(TraitError):
        dumb.date = None
    with pytest.raises(TraitError):
        dumb.date = {'key': 1}  # not allowed
    with pytest.raises(TraitError):
        dumb.date = '1970-01-01T00:00:00.nousec'
    # format it
    dumb.date = '1970-01-01'
    assert resource.DateTime.isoformat(dumb.date).startswith('1970-01-01')


def test_semanticversion_trait():
    class Dummy(HasTraits):
        ver = resource.SemanticVersion()

    dumb = Dummy()

    #  1           => valid    => (1, 0, 0, '')
    for x in '1', (1,), [1,]:
        dumb.ver = x
        assert dumb.ver == (1, 0, 0, '')
    #  rc3         => invalid: no number
    with pytest.raises(TraitError):
        dumb.ver = 'rc3'
    #  1.1         => valid    => (1, 1, 0, '')
    dumb.ver = '1.1'
    assert dumb.ver == (1, 1, 0, '')
    #  1a          => valid    => (1, 0, 0, 'a')
    for x in '1a', (1, 0, 0, 'a'), [1, 0, 0, 'a']:
        dumb.ver = x
        assert dumb.ver == (1, 0, 0, 'a')
    #  1.a.1       => invalid: non-numeric can only go at end
    with pytest.raises(TraitError):
        dumb.ver = '1.a.1'
    #  1.12.1      => valid    => (1, 12, 1, '')
    dumb.ver = '1.12.1'
    assert dumb.ver == (1, 12, 1, '')
    #  1.12.13-1   => valid    => (1, 12, 13, '1')
    dumb.ver = '1.12.13-1a_3'
    assert dumb.ver == (1, 12, 13, '1a_3')
    #  1.12.13.x   => invalid: too many parts
    with pytest.raises(TraitError):
        dumb.ver = '1.12.13.x'
    # invalid last version
    with pytest.raises(TraitError):
        dumb.ver = (1, 2, 3, 3.1415)
    # invalid input length
    with pytest.raises(TraitError):
        dumb.ver = []
    # invalid input type
    with pytest.raises(TraitError):
        dumb.ver = {'not': 'alllowed'}
    # formatting
    dumb.ver = '1.2.3a'
    s = resource.SemanticVersion.pretty(dumb.ver)
    assert s == '1.2.3-a'


def test_identifier_trait():
    class Dummy(HasTraits):
        ver = resource.Identifier()

    dumb = Dummy()
    # non-hex value
    with pytest.raises(TraitError):
        dumb.ver = 'loremipsumdolor--loremipsumdolor'
    # non-string balue
    with pytest.raises(TraitError):
        dumb.ver = 100


def test_traitcontainer_fromdict():
    class Dummy(resource.TraitContainer):
        ident = resource.Identifier()
        date = resource.DateTime()
        tags = List(Unicode())
        meta = Dict()

    # simple initialization
    Dummy.from_dict({})
    # invalid field
    with pytest.raises(KeyError):
        Dummy.from_dict({'surprise': 1})
    # bad list value
    with pytest.raises(ValueError):
        Dummy.from_dict({'tags': 'not a list'})
    # bad dict value
    with pytest.raises(ValueError):
        Dummy.from_dict({'meta': 'not a dict'})


def test_traitcontainer_filepath():
    class Dummy(resource.TraitContainer):
        fp = resource.FilePath()

    dumb = Dummy()
    # test open/read
    dumb.fp.path = __file__
    dumb.fp.open()
    bytes = dumb.fp.read(100)
    # bad init
    with pytest.raises(ValueError):
        dumb.fp = resource.FilePath(tempfile=True, copy=False)


def test_traitcontainer_version():
    class Dummy(resource.TraitContainer):
        ver = resource.Version()

    dumb = Dummy()
    dumb.ver.revision = '1.0.1-alpha'
    dumb.ver.name = 'Foo'
    dumb.ver.created = '1970-01-01'
    assert str(dumb.ver) == 'Foo 1.0.1-alpha (1970-01-01T00:00:00)'


def test_resource_help():
    r = resource.Resource()
    s = r.help('creator')
    assert isinstance(s, str)
    # bad attr
    s = r.help('surprise')
    assert s.startswith('Error') and 'attribute' in s


def test_resource_copy():
    r = resource.Resource(name='foo',
                          version=resource.Version(revision='1.2.3'))
    c = r.copy(version=resource.Version(revision='1.2.4'))
    assert c.name == r.name


def test_resource_table():
    from idaes.dmf import tabular
    # create table at init time
    t = tabular.Table(data=[{
                "name": "Density Data",
                "units": "g/cm^3",
                "values": [1.0053, 1.0188 ],
                "errors": [.00005, 0.00005],
                "error_type": "absolute"
    }])
    r = resource.TabularDataResource(table=t)
    v = r.table.data.values_dataframe()['Density Data']
    assert v[0] == pytest.approx(1.0053)
    # add table in data attr.
    r = resource.Resource(type=resource.ResourceTypes.tabular_data)
    r.data = {
        'data': [
            {
                "name": "Density Data",
                "units": "g/cm^3",
                "values": [1.0053, 1.0188 ],
                "errors": [.00005, 0.00005],
                "error_type": "absolute",
                'type': 'property'
            },
            {
                'name': 'temp',
                'units': 'C',
                'values': [10, 20],
                'type': 'state'
            }
        ]
    }
    v = r.table.data.values_dataframe()['Density Data']
    assert v[0] == pytest.approx(1.0053)





