##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Equilibrium Reactor unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        Holdup0D
from idaes.models.equilibrium_reactor import EquilibriumReactor, \
                                             EquilibriumReactorData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'reaction_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("EquilibriumReactorFrame")
class _EquilibriumReactor(EquilibriumReactorData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.req = EquilibriumReactorFrame(property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.req.config) == 18
    assert not m.fs.req.config.dynamic
    assert not m.fs.req.config.include_holdup
    assert m.fs.req.config.material_balance_type == 'component_phase'
    assert m.fs.req.config.energy_balance_type == 'total'
    assert m.fs.req.config.momentum_balance_type == 'total'
    assert m.fs.req.config.has_rate_reactions
    assert m.fs.req.config.has_equilibrium_reactions
    assert not m.fs.req.config.has_phase_equilibrium
    assert not m.fs.req.config.has_mass_transfer
    assert m.fs.req.config.has_heat_transfer
    assert not m.fs.req.config.has_work_transfer
    assert not m.fs.req.config.has_pressure_change
    assert m.fs.req.config.property_package == m.fs.props
    assert m.fs.req.config.property_package_args == {}
    assert m.fs.req.config.inlet_list is None
    assert m.fs.req.config.num_inlets is None
    assert m.fs.req.config.outlet_list is None
    assert m.fs.req.config.num_outlets is None


def test_super_build(m):
    super(EquilibriumReactorData, m.fs.req).build()
    assert not m.fs.req.config.dynamic


def test_adding_holdup(m):
    super(EquilibriumReactorData, m.fs.req).build()
    m.fs.req.holdup = Holdup0D()

    assert not m.fs.req.holdup.config.include_holdup
    assert m.fs.req.holdup.config.material_balance_type == 'component_phase'
    assert m.fs.req.holdup.config.energy_balance_type == 'total'
    assert m.fs.req.holdup.config.momentum_balance_type == 'total'
    assert m.fs.req.holdup.config.has_rate_reactions
    # Properties do not have equilibrium reactions, so this should be False
    assert not m.fs.req.holdup.config.has_equilibrium_reactions
    assert not m.fs.req.holdup.config.has_phase_equilibrium
    assert not m.fs.req.holdup.config.has_mass_transfer
    assert m.fs.req.holdup.config.has_heat_transfer
    assert not m.fs.req.holdup.config.has_work_transfer
    assert not m.fs.req.holdup.config.has_pressure_change
    assert m.fs.req.holdup.config.property_package == m.fs.props
    assert m.fs.req.holdup.config.property_package_args == {}


def test_set_geometry_include_holdup_false(m):
    super(EquilibriumReactorData, m.fs.req).build()
    m.fs.req.holdup = Holdup0D()
    m.fs.req._set_geometry()

    assert not hasattr(m.fs.req, "volume")


def test_set_geometry_include_holdup_true(m):
    m.fs.req.config.include_holdup = True
    super(EquilibriumReactorData, m.fs.req).build()
    m.fs.req.holdup = Holdup0D()
    m.fs.req._set_geometry()

    assert hasattr(m.fs.req, "volume")
    assert hasattr(m.fs.req.holdup, "volume")
    assert m.fs.req.volume == m.fs.req.holdup.volume


def test_make_performance_rate_reactions_true(m):
    super(EquilibriumReactorData, m.fs.req).build()
    m.fs.req.holdup = Holdup0D()
    m.fs.req._set_geometry()
    m.fs.req._make_performance()

    assert hasattr(m.fs.req, "rate_reaction_idx")
    assert hasattr(m.fs.req, "rate_reaction_equilibrium")


def test_make_performance_rate_reactions_false(m):
    m.fs.req.config.has_rate_reactions = False
    super(EquilibriumReactorData, m.fs.req).build()
    m.fs.req.holdup = Holdup0D()
    m.fs.req._set_geometry()
    m.fs.req._make_performance()

    assert not hasattr(m.fs.req, "rate_reaction_idx")
    assert not hasattr(m.fs.req, "rate_reaction_equilibrium")


def test_make_performance_has_heat_transfer_true(m):
    super(EquilibriumReactorData, m.fs.req).build()
    m.fs.req.holdup = Holdup0D()
    m.fs.req._set_geometry()
    m.fs.req._make_performance()

    assert hasattr(m.fs.req, "heat")


def test_make_performance_has_heat_transfer_false(m):
    m.fs.req.config.has_heat_transfer = False
    super(EquilibriumReactorData, m.fs.req).build()
    m.fs.req.holdup = Holdup0D()
    m.fs.req._set_geometry()
    m.fs.req._make_performance()

    assert not hasattr(m.fs.req, "heat")


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.req = EquilibriumReactor(property_package=m2.fs.props)
    return m2


def test_post_transform_build_SISO(m2):
    m2.fs.req.post_transform_build()

    assert hasattr(m2.fs.req, "inlet")
    assert hasattr(m2.fs.req, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.req.inlet:
        assert k in [0]
    for k in m2.fs.req.outlet:
        assert k in [0]


def test_post_transform_build_inlet_list(m2):
    m2.fs.req.config.inlet_list = ["a", "b"]
    m2.fs.req.post_transform_build()

    assert hasattr(m2.fs.req, "inlet")
    assert hasattr(m2.fs.req, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.req.inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.req.outlet:
        assert k in [0]


def test_post_transform_build_num_inlets(m2):
    m2.fs.req.config.num_inlets = 2
    m2.fs.req.post_transform_build()

    assert hasattr(m2.fs.req, "inlet")
    assert hasattr(m2.fs.req, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.req.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.req.outlet:
        assert k in [0]


def test_post_transform_build_outlet_list(m2):
    m2.fs.req.config.outlet_list = ["a", "b"]
    m2.fs.req.post_transform_build()

    assert hasattr(m2.fs.req, "inlet")
    assert hasattr(m2.fs.req, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.req.inlet:
        assert k in [0]
    for k in m2.fs.req.outlet:
        assert k in [(0, "a"), (0, "b")]


def test_post_transform_build_num_outlet2(m2):
    m2.fs.req.config.num_outlets = 2
    m2.fs.req.post_transform_build()

    assert hasattr(m2.fs.req, "inlet")
    assert hasattr(m2.fs.req, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.req.inlet:
        assert k in [0]
    for k in m2.fs.req.outlet:
        assert k in [(0, "1"), (0, "2")]


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.req = EquilibriumReactor(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.req.inlet[:].vars["flow_mol_comp"]["a"].fix(1.0)
    m.fs.req.inlet[:].vars["flow_mol_comp"]["b"].fix(2.0)
    m.fs.req.inlet[:].vars["flow_mol_comp"]["c"].fix(0.1)
    m.fs.req.inlet[:].vars["flow_mol_comp"]["d"].fix(0.0)
    m.fs.req.inlet[:].vars["flow_mol_comp"]["e"].fix(0.0)
    m.fs.req.inlet[:].vars["flow_mol_comp"]["f"].fix(0.0)
    m.fs.req.inlet[:].vars["enth_mol"].fix(-238.71)
    m.fs.req.inlet[:].vars["pressure"].fix(101325.0)

    m.fs.req.heat.fix(0.0)

    assert degrees_of_freedom(m) == 0

    m.fs.req.initialize()

    assert (pytest.approx(0.058, abs=1e-2) ==
            m.fs.req.outlet[0].vars["flow_mol_comp"]["a"].value)
    assert (pytest.approx(0.539, abs=1e-2) ==
            m.fs.req.outlet[0].vars["flow_mol_comp"]["b"].value)
    assert (pytest.approx(0.599, abs=1e-2) ==
            m.fs.req.outlet[0].vars["flow_mol_comp"]["c"].value)
    assert (pytest.approx(0.540, abs=1e-2) ==
            m.fs.req.outlet[0].vars["flow_mol_comp"]["d"].value)
    assert (pytest.approx(0.041, abs=1e-2) ==
            m.fs.req.outlet[0].vars["flow_mol_comp"]["e"].value)
    assert (pytest.approx(0.382, abs=1e-2) ==
            m.fs.req.outlet[0].vars["flow_mol_comp"]["f"].value)
    assert (pytest.approx(-342.871, abs=1e-2) ==
            m.fs.req.outlet[0].vars["enth_mol"].value)
    assert (pytest.approx(101325.0, abs=1e-2) ==
            m.fs.req.outlet[0].vars["pressure"].value)
