##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Flash unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        Holdup0D
from idaes.models.flash import Flash, FlashData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'phase_equilibrium_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("FlashFrame")
class _Flash(FlashData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.flash = FlashFrame(property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.flash.config) == 18
    assert m.fs.flash.config.dynamic == 'use_parent_value'
    assert not m.fs.flash.config.include_holdup
    assert m.fs.flash.config.material_balance_type == 'component_phase'
    assert m.fs.flash.config.energy_balance_type == 'total'
    assert m.fs.flash.config.momentum_balance_type == 'total'
    assert not m.fs.flash.config.has_rate_reactions
    assert not m.fs.flash.config.has_equilibrium_reactions
    assert m.fs.flash.config.has_phase_equilibrium
    assert not m.fs.flash.config.has_mass_transfer
    assert m.fs.flash.config.has_heat_transfer
    assert not m.fs.flash.config.has_work_transfer
    assert m.fs.flash.config.has_pressure_change
    assert m.fs.flash.config.property_package == m.fs.props
    assert m.fs.flash.config.property_package_args == {}
    assert m.fs.flash.config.inlet_list is None
    assert m.fs.flash.config.num_inlets is None
    assert m.fs.flash.config.outlet_list == ["Vap", "Liq"]
    assert m.fs.flash.config.num_outlets is None


def test_super_build(m):
    super(FlashData, m.fs.flash).build()
    assert not m.fs.flash.config.dynamic


def test_adding_holdup(m):
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()

    assert not m.fs.flash.holdup.config.include_holdup
    assert m.fs.flash.holdup.config.material_balance_type == 'component_phase'
    assert m.fs.flash.holdup.config.energy_balance_type == 'total'
    assert m.fs.flash.holdup.config.momentum_balance_type == 'total'
    assert not m.fs.flash.holdup.config.has_rate_reactions
    assert not m.fs.flash.holdup.config.has_equilibrium_reactions
    assert m.fs.flash.holdup.config.has_phase_equilibrium
    assert not m.fs.flash.holdup.config.has_mass_transfer
    assert m.fs.flash.holdup.config.has_heat_transfer
    assert not m.fs.flash.holdup.config.has_work_transfer
    assert m.fs.flash.holdup.config.has_pressure_change
    assert m.fs.flash.holdup.config.property_package == m.fs.props
    assert m.fs.flash.holdup.config.property_package_args == {}


def test_set_geometry_include_holdup_false(m):
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()

    assert not hasattr(m.fs.flash, "volume")


def test_set_geometry_include_holdup_true(m):
    m.fs.flash.config.include_holdup = True
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()

    assert hasattr(m.fs.flash, "volume")
    assert hasattr(m.fs.flash.holdup, "volume")
    assert m.fs.flash.volume == m.fs.flash.holdup.volume


def test_make_performance_has_heat_transfer_true(m):
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()
    m.fs.flash.build_outlets(outlets=m.fs.flash.config.outlet_list,
                             num_outlets=m.fs.flash.config.num_outlets,
                             split_type="phase")
    m.fs.flash._make_performance()

    assert hasattr(m.fs.flash, "heat")


def test_make_performance_has_heat_transfer_false(m):
    m.fs.flash.config.has_heat_transfer = False
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()
    m.fs.flash.build_outlets(outlets=m.fs.flash.config.outlet_list,
                             num_outlets=m.fs.flash.config.num_outlets,
                             split_type="phase")
    m.fs.flash._make_performance()

    assert not hasattr(m.fs.flash, "heat")


def test_make_performance_has_pressure_change_true(m):
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()
    m.fs.flash.build_outlets(outlets=m.fs.flash.config.outlet_list,
                             num_outlets=m.fs.flash.config.num_outlets,
                             split_type="phase")
    m.fs.flash._make_performance()

    assert hasattr(m.fs.flash, "deltaP")


def test_make_performance_has_pressure_change_false(m):
    m.fs.flash.config.has_pressure_change = False
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()
    m.fs.flash.build_outlets(outlets=m.fs.flash.config.outlet_list,
                             num_outlets=m.fs.flash.config.num_outlets,
                             split_type="phase")
    m.fs.flash._make_performance()

    assert not hasattr(m.fs.flash, "deltaP")

# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.flash = Flash(property_package=m2.fs.props)
    return m2


def test_post_transform_build_no_args(m2):
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [0]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "Vap"), (0, "Liq")]

    # Check that split fractions have been set correctly
    assert m2.fs.flash.holdup.outlet_splitter.split_fraction[0,
                                                             "Vap",
                                                             "Vap"].fixed
    assert m2.fs.flash.holdup.outlet_splitter.split_fraction[0,
                                                             "Vap",
                                                             "Vap"].value == 1
    assert m2.fs.flash.holdup.outlet_splitter.split_fraction[0,
                                                             "Liq",
                                                             "Liq"].fixed
    assert m2.fs.flash.holdup.outlet_splitter.split_fraction[0,
                                                             "Liq",
                                                             "Liq"].value == 1


def test_post_transform_build_inlet_list(m2):
    m2.fs.flash.config.inlet_list = ["a", "b"]
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "Vap"), (0, "Liq")]


def test_post_transform_build_num_inlets(m2):
    m2.fs.flash.config.num_inlets = 2
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "Vap"), (0, "Liq")]


def test_post_transform_build_outlet_list(m2):
    m2.fs.flash.config.outlet_list = ["a", "b"]
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [0]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "a"), (0, "b")]

    # Check that split fractions have not been fixed
    for k in m2.fs.flash.holdup.outlet_splitter.split_fraction[0, :, :]:
        assert not k.fixed


def test_post_transform_build_num_outlet2(m2):
    m2.fs.flash.config.outlet_list = None
    m2.fs.flash.config.num_outlets = 2
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [0]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "1"), (0, "2")]

    # Check that split fractions have not been fixed
    for k in m2.fs.flash.holdup.outlet_splitter.split_fraction[0, :, :]:
        assert not k.fixed

# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.flash = Flash(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.flash.inlet[:].vars["flow_mol_comp_vap"]["CO2"].fix(0.1)
    m.fs.flash.inlet[:].vars["flow_mol_comp_vap"]["H2O"].fix(0.1)
    m.fs.flash.inlet[:].vars["flow_mol_comp_vap"]["N2"].fix(0.8)
    m.fs.flash.inlet[:].vars["flow_mol_comp_liq"]["CO2"].fix(0.0)
    m.fs.flash.inlet[:].vars["flow_mol_comp_liq"]["H2O"].fix(2.0)
    m.fs.flash.inlet[:].vars["temperature"].fix(300.0)
    m.fs.flash.inlet[:].vars["pressure"].fix(101325.0)

    m.fs.flash.heat.fix(0.0)
    m.fs.flash.deltaP.fix(0.0)

    assert degrees_of_freedom(m) == 0

    m.fs.flash.initialize()

    assert (
        pytest.approx(0.000328, abs=1e-5) ==
        m.fs.flash.outlet[0, "Liq"].vars["flow_mol_comp_liq"]["CO2"].value)
    assert (
        pytest.approx(2.068, abs=1e-2) ==
        m.fs.flash.outlet[0, "Liq"].vars["flow_mol_comp_liq"]["H2O"].value)
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Liq"].vars["flow_mol_comp_vap"]["CO2"].value)
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Liq"].vars["flow_mol_comp_vap"]["H2O"].value)
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Liq"].vars["flow_mol_comp_vap"]["N2"].value)
    assert (pytest.approx(300.056, abs=1e-2) ==
            m.fs.flash.outlet[0, "Liq"].vars["temperature"].value)
    assert (pytest.approx(101325.0, abs=1e-2) ==
            m.fs.flash.outlet[0, "Liq"].vars["pressure"].value)

    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Vap"].vars["flow_mol_comp_liq"]["CO2"].value)
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Vap"].vars["flow_mol_comp_liq"]["H2O"].value)
    assert (
        pytest.approx(0.100, abs=1e-2) ==
        m.fs.flash.outlet[0, "Vap"].vars["flow_mol_comp_vap"]["CO2"].value)
    assert (
        pytest.approx(0.0323, abs=1e-3) ==
        m.fs.flash.outlet[0, "Vap"].vars["flow_mol_comp_vap"]["H2O"].value)
    assert (
        pytest.approx(0.8, abs=1e-2) ==
        m.fs.flash.outlet[0, "Vap"].vars["flow_mol_comp_vap"]["N2"].value)
    assert (pytest.approx(300.056, abs=1e-2) ==
            m.fs.flash.outlet[0, "Vap"].vars["temperature"].value)
    assert (pytest.approx(101325.0, abs=1e-2) ==
            m.fs.flash.outlet[0, "Vap"].vars["pressure"].value)
