##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Heat Exchanger unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        Holdup1D
from idaes.models.heat_exchanger_1D import HeatExchanger1D, HeatExchanger1dData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'steam_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("HeatExchanger1dFrame")
class _HeatExchanger1D(HeatExchanger1dData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.hx = HeatExchanger1dFrame(shell_property_package=m.fs.props,
                                   tube_property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.hx.config) == 35
    assert m.fs.hx.config.dynamic == 'use_parent_value'
    assert not m.fs.hx.config.include_holdup
    assert m.fs.hx.config.material_balance_type == 'component_phase'
    assert m.fs.hx.config.energy_balance_type == 'total'
    assert m.fs.hx.config.momentum_balance_type == 'total'
    assert not m.fs.hx.config.has_rate_reactions
    assert not m.fs.hx.config.has_equilibrium_reactions
    assert not m.fs.hx.config.has_mass_transfer
    assert m.fs.hx.config.has_heat_transfer
    assert not m.fs.hx.config.has_work_transfer
    assert not m.fs.hx.config.has_pressure_change

    assert not hasattr(m.fs.hx.config, "has_phase_equilibrium")

    assert m.fs.hx.config.shell_has_phase_equilibrium
    m.fs.hx.config.shell_has_phase_equilibrium = False
    with pytest.raises(ValueError):
        m.fs.hx.config.shell_has_phase_equilibrium = 'foo'
        m.fs.hx.config.shell_has_phase_equilibrium = 10

    assert m.fs.hx.config.tube_has_phase_equilibrium
    m.fs.hx.config.tube_has_phase_equilibrium = False
    with pytest.raises(ValueError):
        m.fs.hx.config.tube_has_phase_equilibrium = 'foo'
        m.fs.hx.config.tube_has_phase_equilibrium = 10

    assert m.fs.hx.config.shell_property_package == m.fs.props
    assert m.fs.hx.config.shell_property_package_args == {}
    assert m.fs.hx.config.shell_inlet_list is None
    assert m.fs.hx.config.shell_num_inlets is None
    assert m.fs.hx.config.shell_outlet_list is None
    assert m.fs.hx.config.shell_num_outlets is None

    assert m.fs.hx.config.tube_property_package == m.fs.props
    assert m.fs.hx.config.tube_property_package_args == {}
    assert m.fs.hx.config.tube_inlet_list is None
    assert m.fs.hx.config.tube_num_inlets is None
    assert m.fs.hx.config.tube_outlet_list is None
    assert m.fs.hx.config.tube_num_outlets is None

    assert m.fs.hx.config.shell_discretization_method == 'OCLR'
    assert m.fs.hx.config.tube_discretization_method == 'OCLR'
    assert m.fs.hx.config.finite_elements == 20
    assert m.fs.hx.config.collocation_points == 3

    assert not m.fs.hx.config.shell_has_mass_diffusion
    m.fs.hx.config.shell_has_mass_diffusion = True
    with pytest.raises(ValueError):
        m.fs.hx.config.shell_has_mass_diffusion = 'foo'
    with pytest.raises(ValueError):
        m.fs.hx.config.shell_has_mass_diffusion = 10

    assert not m.fs.hx.config.shell_has_energy_diffusion
    m.fs.hx.config.shell_has_energy_diffusion = True
    with pytest.raises(ValueError):
        m.fs.hx.config.shell_has_energy_diffusion = 'foo'
    with pytest.raises(ValueError):
        m.fs.hx.config.shell_has_energy_diffusion = 10

    assert not m.fs.hx.config.tube_has_mass_diffusion
    m.fs.hx.config.tube_has_mass_diffusion = True
    with pytest.raises(ValueError):
        m.fs.hx.config.tube_has_mass_diffusion = 'foo'
    with pytest.raises(ValueError):
        m.fs.hx.config.tube_has_mass_diffusion = 10

    assert not m.fs.hx.config.tube_has_energy_diffusion
    m.fs.hx.config.tube_has_energy_diffusion = True
    with pytest.raises(ValueError):
        m.fs.hx.config.tube_has_energy_diffusion = 'foo'
    with pytest.raises(ValueError):
        m.fs.hx.config.tube_has_energy_diffusion = 10

    assert m.fs.hx.config.flow_type == 'co_current'

    assert m.fs.hx.config.has_wall_conduction == 'none'


def test_super_build(m):
    super(HeatExchanger1dData, m.fs.hx).build()
    assert not m.fs.hx.config.dynamic


def test_adding_holdup(m):
    super(HeatExchanger1dData, m.fs.hx).build()
    m.fs.hx.shell = Holdup1D(
            property_package=m.fs.hx.config.shell_property_package,
            property_package_args=m.fs.hx.config.shell_property_package_args,
            flow_direction='forward',
            discretization_method=m.fs.hx.config.shell_discretization_method,
            finite_elements=m.fs.hx.config.finite_elements,
            collocation_points=m.fs.hx.config.collocation_points)

    m.fs.hx.tube = Holdup1D(
            property_package=m.fs.hx.config.tube_property_package,
            property_package_args=m.fs.hx.config.tube_property_package_args,
            flow_direction='forward',
            discretization_method=m.fs.hx.config.tube_discretization_method,
            finite_elements=m.fs.hx.config.finite_elements,
            collocation_points=m.fs.hx.config.collocation_points)

    assert not m.fs.hx.shell.config.include_holdup
    assert m.fs.hx.shell.config.material_balance_type == 'component_phase'
    assert m.fs.hx.shell.config.energy_balance_type == 'total'
    assert m.fs.hx.shell.config.momentum_balance_type == 'total'
    assert not m.fs.hx.shell.config.has_rate_reactions
    assert not m.fs.hx.shell.config.has_equilibrium_reactions
    # Properties do not have phase equilibrium , so this should be False
    assert not m.fs.hx.shell.config.has_phase_equilibrium
    assert not m.fs.hx.shell.config.has_mass_transfer
    assert m.fs.hx.shell.config.has_heat_transfer
    assert not m.fs.hx.shell.config.has_work_transfer
    assert not m.fs.hx.shell.config.has_pressure_change
    assert m.fs.hx.shell.config.property_package == m.fs.props
    assert m.fs.hx.shell.config.property_package_args == {}

    assert not m.fs.hx.tube.config.include_holdup
    assert m.fs.hx.tube.config.material_balance_type == 'component_phase'
    assert m.fs.hx.tube.config.energy_balance_type == 'total'
    assert m.fs.hx.tube.config.momentum_balance_type == 'total'
    assert not m.fs.hx.tube.config.has_rate_reactions
    assert not m.fs.hx.tube.config.has_equilibrium_reactions
    # Properties do not have phase equilibrium , so this should be False
    assert not m.fs.hx.tube.config.has_phase_equilibrium
    assert not m.fs.hx.tube.config.has_mass_transfer
    assert m.fs.hx.tube.config.has_heat_transfer
    assert not m.fs.hx.tube.config.has_work_transfer
    assert not m.fs.hx.tube.config.has_pressure_change
    assert m.fs.hx.tube.config.property_package == m.fs.props
    assert m.fs.hx.tube.config.property_package_args == {}


def test_set_geometry(m):
    super(HeatExchanger1dData, m.fs.hx).build()
    m.fs.hx.shell = Holdup1D(
            property_package=m.fs.hx.config.shell_property_package,
            property_package_args=m.fs.hx.config.shell_property_package_args,
            flow_direction='forward',
            discretization_method=m.fs.hx.config.shell_discretization_method,
            finite_elements=m.fs.hx.config.finite_elements,
            collocation_points=m.fs.hx.config.collocation_points)

    m.fs.hx.tube = Holdup1D(
            property_package=m.fs.hx.config.tube_property_package,
            property_package_args=m.fs.hx.config.tube_property_package_args,
            flow_direction='forward',
            discretization_method=m.fs.hx.config.tube_discretization_method,
            finite_elements=m.fs.hx.config.finite_elements,
            collocation_points=m.fs.hx.config.collocation_points)
    m.fs.hx._set_geometry()

    assert hasattr(m.fs.hx, "shell_area")
    assert hasattr(m.fs.hx, "shell_length")
    assert hasattr(m.fs.hx, "tube_area")
    assert hasattr(m.fs.hx, "tube_length")


def test_make_performance(m):
    super(HeatExchanger1dData, m.fs.hx).build()
    m.fs.hx.shell = Holdup1D(
            property_package=m.fs.hx.config.shell_property_package,
            property_package_args=m.fs.hx.config.shell_property_package_args,
            flow_direction='forward',
            discretization_method=m.fs.hx.config.shell_discretization_method,
            finite_elements=m.fs.hx.config.finite_elements,
            collocation_points=m.fs.hx.config.collocation_points)

    m.fs.hx.tube = Holdup1D(
            property_package=m.fs.hx.config.tube_property_package,
            property_package_args=m.fs.hx.config.tube_property_package_args,
            flow_direction='forward',
            discretization_method=m.fs.hx.config.tube_discretization_method,
            finite_elements=m.fs.hx.config.finite_elements,
            collocation_points=m.fs.hx.config.collocation_points)
    m.fs.hx._set_geometry()
    m.fs.hx._make_performance()

    assert hasattr(m.fs.hx, "pi")
    assert hasattr(m.fs.hx, "d_shell")
    assert hasattr(m.fs.hx, "d_tube_outer")
    assert hasattr(m.fs.hx, "d_tube_inner")
    assert hasattr(m.fs.hx, "N_tubes")
    assert hasattr(m.fs.hx, "temperature_wall")
    assert hasattr(m.fs.hx, "shell_heat_transfer_coefficient")
    assert hasattr(m.fs.hx, "tube_heat_transfer_coefficient")
    assert hasattr(m.fs.hx, "shell_heat_transfer_eq")
    assert hasattr(m.fs.hx, "tube_heat_transfer_eq")
    assert hasattr(m.fs.hx, "wall_0D_model")
    assert hasattr(m.fs.hx, "area_calc_tube")
    assert hasattr(m.fs.hx, "area_calc_shell")


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.hx = HeatExchanger1D(shell_property_package=m2.fs.props,
                               tube_property_package=m2.fs.props)
    return m2


def test_post_transform_build_no_args(m2):
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "shell_inlet")
    assert hasattr(m2.fs.hx, "shell_outlet")
    assert hasattr(m2.fs.hx, "tube_inlet")
    assert hasattr(m2.fs.hx, "tube_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.shell_inlet:
        assert k in [0]
    for k in m2.fs.hx.shell_outlet:
        assert k in [0]
    for k in m2.fs.hx.tube_inlet:
        assert k in [0]
    for k in m2.fs.hx.tube_outlet:
        assert k in [0]


def test_post_transform_build_shell_inlet_list(m2):
    m2.fs.hx.config.shell_inlet_list = ["a", "b"]
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "shell_inlet")
    assert hasattr(m2.fs.hx, "shell_outlet")
    assert hasattr(m2.fs.hx, "tube_inlet")
    assert hasattr(m2.fs.hx, "tube_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.shell_inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.hx.shell_outlet:
        assert k in [0]
    for k in m2.fs.hx.tube_inlet:
        assert k in [0]
    for k in m2.fs.hx.tube_outlet:
        assert k in [0]


def test_post_transform_build_shell_num_inlets(m2):
    m2.fs.hx.config.shell_num_inlets = 2
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "shell_inlet")
    assert hasattr(m2.fs.hx, "shell_outlet")
    assert hasattr(m2.fs.hx, "tube_inlet")
    assert hasattr(m2.fs.hx, "tube_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.shell_inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.hx.shell_outlet:
        assert k in [0]
    for k in m2.fs.hx.tube_inlet:
        assert k in [0]
    for k in m2.fs.hx.tube_outlet:
        assert k in [0]


def test_post_transform_build_tube_inlet_list(m2):
    m2.fs.hx.config.tube_inlet_list = ["a", "b"]
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "shell_inlet")
    assert hasattr(m2.fs.hx, "shell_outlet")
    assert hasattr(m2.fs.hx, "tube_inlet")
    assert hasattr(m2.fs.hx, "tube_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.shell_inlet:
        assert k in [0]
    for k in m2.fs.hx.shell_outlet:
        assert k in [0]
    for k in m2.fs.hx.tube_inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.hx.tube_outlet:
        assert k in [0]


def test_post_transform_build_tube_num_inlets(m2):
    m2.fs.hx.config.tube_num_inlets = 2
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "shell_inlet")
    assert hasattr(m2.fs.hx, "shell_outlet")
    assert hasattr(m2.fs.hx, "tube_inlet")
    assert hasattr(m2.fs.hx, "tube_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.shell_inlet:
        assert k in [0]
    for k in m2.fs.hx.shell_outlet:
        assert k in [0]
    for k in m2.fs.hx.tube_inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.hx.tube_outlet:
        assert k in [0]


def test_post_transform_build_shell_outlet_list(m2):
    m2.fs.hx.config.shell_outlet_list = ["a", "b"]
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "shell_inlet")
    assert hasattr(m2.fs.hx, "shell_outlet")
    assert hasattr(m2.fs.hx, "tube_inlet")
    assert hasattr(m2.fs.hx, "tube_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.shell_inlet:
        assert k in [0]
    for k in m2.fs.hx.shell_outlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.hx.tube_inlet:
        assert k in [0]
    for k in m2.fs.hx.tube_outlet:
        assert k in [0]


def test_post_transform_build_shell_num_outlet2(m2):
    m2.fs.hx.config.shell_num_outlets = 2
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "shell_inlet")
    assert hasattr(m2.fs.hx, "shell_outlet")
    assert hasattr(m2.fs.hx, "tube_inlet")
    assert hasattr(m2.fs.hx, "tube_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.shell_inlet:
        assert k in [0]
    for k in m2.fs.hx.shell_outlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.hx.tube_inlet:
        assert k in [0]
    for k in m2.fs.hx.tube_outlet:
        assert k in [0]


def test_post_transform_build_tube_outlet_list(m2):
    m2.fs.hx.config.tube_outlet_list = ["a", "b"]
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "shell_inlet")
    assert hasattr(m2.fs.hx, "shell_outlet")
    assert hasattr(m2.fs.hx, "tube_inlet")
    assert hasattr(m2.fs.hx, "tube_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.shell_inlet:
        assert k in [0]
    for k in m2.fs.hx.shell_outlet:
        assert k in [0]
    for k in m2.fs.hx.tube_inlet:
        assert k in [0]
    for k in m2.fs.hx.tube_outlet:
        assert k in [(0, "a"), (0, "b")]


def test_post_transform_build_tube_num_outlet2(m2):
    m2.fs.hx.config.tube_num_outlets = 2
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "shell_inlet")
    assert hasattr(m2.fs.hx, "shell_outlet")
    assert hasattr(m2.fs.hx, "tube_inlet")
    assert hasattr(m2.fs.hx, "tube_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.shell_inlet:
        assert k in [0]
    for k in m2.fs.hx.shell_outlet:
        assert k in [0]
    for k in m2.fs.hx.tube_inlet:
        assert k in [0]
    for k in m2.fs.hx.tube_outlet:
        assert k in [(0, "1"), (0, "2")]


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.hx = HeatExchanger1D(shell_property_package=m.fs.props,
                              tube_property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.hx.shell_inlet[:].vars["flow_mol_comp"]["H2O"].fix(27.5)
    m.fs.hx.shell_inlet[:].vars["temperature"].fix(866.5)
    m.fs.hx.shell_inlet[:].vars["pressure"].fix(2.97e7)

    m.fs.hx.tube_inlet[:].vars["flow_mol_comp"]["H2O"].fix(27.5)
    m.fs.hx.tube_inlet[:].vars["temperature"].fix(500.0)
    m.fs.hx.tube_inlet[:].vars["pressure"].fix(2.97e7)

    m.fs.hx.d_shell.fix(1.0)
    m.fs.hx.d_tube_outer.fix(0.5)
    m.fs.hx.d_tube_inner.fix(0.499)
    m.fs.hx.N_tubes.fix(1)
    m.fs.hx.shell_length.fix(1.0)
    m.fs.hx.tube_length.fix(1.0)
    m.fs.hx.shell_heat_transfer_coefficient.fix(1000.0)
    m.fs.hx.tube_heat_transfer_coefficient.fix(500.0)

    assert degrees_of_freedom(m) == 0

    m.fs.hx.initialize()

    assert (pytest.approx(27.5, abs=1e-2) ==
            m.fs.hx.shell_outlet[0].vars["flow_mol_comp"]["H2O"].value)
    assert (pytest.approx(787.407, abs=1e-2) ==
            m.fs.hx.shell_outlet[0].vars["temperature"].value)
    assert (pytest.approx(29700000.0, abs=1e-2) ==
            m.fs.hx.shell_outlet[0].vars["pressure"].value)

    assert (pytest.approx(27.5, abs=1e-2) ==
            m.fs.hx.tube_outlet[0].vars["flow_mol_comp"]["H2O"].value)
    assert (pytest.approx(530.54, abs=1e-2) ==
            m.fs.hx.tube_outlet[0].vars["temperature"].value)
    assert (pytest.approx(29700000.0, abs=1e-2) ==
            m.fs.hx.tube_outlet[0].vars["pressure"].value)
