##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Product unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        HoldupStatic
from idaes.models.product import Product, ProductData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'steam_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("ProductFrame")
class _Product(ProductData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.product = ProductFrame(property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.product.config) == 16
    assert not m.fs.product.config.dynamic
    assert not m.fs.product.config.include_holdup
    assert m.fs.product.config.material_balance_type == 'component_phase'
    assert m.fs.product.config.energy_balance_type == 'total'
    assert m.fs.product.config.momentum_balance_type == 'total'
    assert not m.fs.product.config.has_rate_reactions
    assert not m.fs.product.config.has_equilibrium_reactions
    assert not m.fs.product.config.has_phase_equilibrium
    assert not m.fs.product.config.has_mass_transfer
    assert not m.fs.product.config.has_heat_transfer
    assert not m.fs.product.config.has_work_transfer
    assert not m.fs.product.config.has_pressure_change
    assert m.fs.product.config.property_package == m.fs.props
    assert m.fs.product.config.property_package_args == {}
    assert m.fs.product.config.inlet_list is None
    assert m.fs.product.config.num_inlets is None


def test_super_build(m):
    super(ProductData, m.fs.product).build()
    assert not m.fs.product.config.dynamic


def test_adding_holdup(m):
    super(ProductData, m.fs.product).build()
    m.fs.product.holdup = HoldupStatic()

    assert not m.fs.product.holdup.config.include_holdup
    assert m.fs.product.holdup.config.material_balance_type == 'component_phase'
    assert m.fs.product.holdup.config.energy_balance_type == 'total'
    assert m.fs.product.holdup.config.momentum_balance_type == 'total'
    assert not m.fs.product.holdup.config.has_rate_reactions
    assert not m.fs.product.holdup.config.has_equilibrium_reactions
    assert not m.fs.product.holdup.config.has_phase_equilibrium
    assert not m.fs.product.holdup.config.has_mass_transfer
    assert not m.fs.product.holdup.config.has_heat_transfer
    assert not m.fs.product.holdup.config.has_work_transfer
    assert not m.fs.product.holdup.config.has_pressure_change
    assert m.fs.product.holdup.config.property_package == m.fs.props
    assert m.fs.product.holdup.config.property_package_args == {}


def test_set_product_references(m):
    super(ProductData, m.fs.product).build()
    m.fs.product.holdup = HoldupStatic()
    m.fs.product._set_product_references()

    assert hasattr(m.fs.product, "product")
    for k in m.fs.product.product[0].vars.keys():
        assert k in ["flow_mol_comp", "pressure", "temperature"]


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.product = Product(property_package=m2.fs.props)
    return m2


def test_post_transform_build_SI(m2):
    m2.fs.product.post_transform_build()

    assert hasattr(m2.fs.product, "inlet")

    # Test that inlet has correct indexing sets
    for k in m2.fs.product.inlet:
        assert k in [0]


def test_post_transform_build_inlet_list(m2):
    m2.fs.product.config.inlet_list = ["a", "b"]
    m2.fs.product.post_transform_build()

    assert hasattr(m2.fs.product, "inlet")

    # Test that inlet has correct indexing sets
    for k in m2.fs.product.inlet:
        assert k in [(0, "a"), (0, "b")]


def test_post_transform_build_num_inlets(m2):
    m2.fs.product.config.num_inlets = 2
    m2.fs.product.post_transform_build()

    assert hasattr(m2.fs.product, "inlet")

    # Test that inlet has correct indexing sets
    for k in m2.fs.product.inlet:
        assert k in [(0, "1"), (0, "2")]


# -----------------------------------------------------------------------------
# Test user interface methods
def test_display():
    # Test that no errors occur when using display
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.product = Product(property_package=m.fs.props)
    m.fs.post_transform_build()
    m.fs.product.display()


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.product = Product(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.product.inlet[:].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
    m.fs.product.inlet[:].vars["temperature"].fix(866.5)
    m.fs.product.inlet[:].vars["pressure"].fix(2.97e7)

    assert degrees_of_freedom(m) == 0

    m.fs.product.initialize()

    assert (pytest.approx(27500.0, abs=1e-2) ==
            m.fs.product.product[0].vars["flow_mol_comp"]["H2O"].value)
    assert (pytest.approx(866.5, abs=1e-2) ==
            m.fs.product.product[0].vars["temperature"].value)
    assert (pytest.approx(29700000.0, abs=1e-2) ==
            m.fs.product.product[0].vars["pressure"].value)
