##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for TemperatureChanger unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        Holdup0D
from idaes.models.temperature_changer import TemperatureChanger, \
                                             TemperatureChangerData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'steam_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("TemperatureChangerFrame")
class _TemperatureChanger(TemperatureChangerData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.tc = TemperatureChangerFrame(property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.tc.config) == 19
    assert m.fs.tc.config.dynamic == 'use_parent_value'
    assert not m.fs.tc.config.include_holdup
    assert m.fs.tc.config.material_balance_type == 'component_phase'
    assert m.fs.tc.config.energy_balance_type == 'total'
    assert m.fs.tc.config.momentum_balance_type == 'total'
    assert not m.fs.tc.config.has_rate_reactions
    assert not m.fs.tc.config.has_equilibrium_reactions
    assert m.fs.tc.config.has_phase_equilibrium
    assert not m.fs.tc.config.has_mass_transfer
    assert m.fs.tc.config.has_heat_transfer
    assert not m.fs.tc.config.has_work_transfer
    assert not m.fs.tc.config.has_pressure_change
    assert m.fs.tc.config.property_package == m.fs.props
    assert m.fs.tc.config.property_package_args == {}
    assert m.fs.tc.config.inlet_list is None
    assert m.fs.tc.config.num_inlets is None
    assert m.fs.tc.config.outlet_list is None
    assert m.fs.tc.config.num_outlets is None

    assert m.fs.tc.config.heater
    m.fs.tc.config.heater = True
    m.fs.tc.config.heater = False
    with pytest.raises(ValueError):
        m.fs.tc.config.heater = 'foo'
    with pytest.raises(ValueError):
        m.fs.tc.config.heater = 10


def test_super_build(m):
    super(TemperatureChangerData, m.fs.tc).build()
    assert not m.fs.tc.config.dynamic


def test_adding_holdup(m):
    super(TemperatureChangerData, m.fs.tc).build()
    m.fs.tc.holdup = Holdup0D()

    assert not m.fs.tc.holdup.config.include_holdup
    assert m.fs.tc.holdup.config.material_balance_type == 'component_phase'
    assert m.fs.tc.holdup.config.energy_balance_type == 'total'
    assert m.fs.tc.holdup.config.momentum_balance_type == 'total'
    assert not m.fs.tc.holdup.config.has_rate_reactions
    # Properties do not have equilibrium reactions, so this should be False
    assert not m.fs.tc.holdup.config.has_equilibrium_reactions
    assert not m.fs.tc.holdup.config.has_phase_equilibrium
    assert not m.fs.tc.holdup.config.has_mass_transfer
    assert m.fs.tc.holdup.config.has_heat_transfer
    assert not m.fs.tc.holdup.config.has_work_transfer
    assert not m.fs.tc.holdup.config.has_pressure_change
    assert m.fs.tc.holdup.config.property_package == m.fs.props
    assert m.fs.tc.holdup.config.property_package_args == {}


def test_set_geometry_include_holdup_false(m):
    super(TemperatureChangerData, m.fs.tc).build()
    m.fs.tc.holdup = Holdup0D()
    m.fs.tc._set_geometry()

    assert not hasattr(m.fs.tc, "volume")


def test_set_geometry_include_holdup_true(m):
    m.fs.tc.config.include_holdup = True
    super(TemperatureChangerData, m.fs.tc).build()
    m.fs.tc.holdup = Holdup0D()
    m.fs.tc._set_geometry()

    assert hasattr(m.fs.tc, "volume")
    assert hasattr(m.fs.tc.holdup, "volume")
    assert m.fs.tc.volume == m.fs.tc.holdup.volume


def test_make_performance(m):
    super(TemperatureChangerData, m.fs.tc).build()
    m.fs.tc.holdup = Holdup0D()
    m.fs.tc._set_geometry()
    m.fs.tc._make_performance()

    assert hasattr(m.fs.tc, "heat")
    assert hasattr(m.fs.tc, "deltaT")
    assert hasattr(m.fs.tc, "temperature_difference")


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.tc = TemperatureChanger(property_package=m2.fs.props)
    return m2


def test_post_transform_build_SISO(m2):
    m2.fs.tc.post_transform_build()

    assert hasattr(m2.fs.tc, "inlet")
    assert hasattr(m2.fs.tc, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.tc.inlet:
        assert k in [0]
    for k in m2.fs.tc.outlet:
        assert k in [0]


def test_post_transform_build_inlet_list(m2):
    m2.fs.tc.config.inlet_list = ["a", "b"]
    m2.fs.tc.post_transform_build()

    assert hasattr(m2.fs.tc, "inlet")
    assert hasattr(m2.fs.tc, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.tc.inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.tc.outlet:
        assert k in [0]


def test_post_transform_build_num_inlets(m2):
    m2.fs.tc.config.num_inlets = 2
    m2.fs.tc.post_transform_build()

    assert hasattr(m2.fs.tc, "inlet")
    assert hasattr(m2.fs.tc, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.tc.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.tc.outlet:
        assert k in [0]


def test_post_transform_build_outlet_list(m2):
    m2.fs.tc.config.outlet_list = ["a", "b"]
    m2.fs.tc.post_transform_build()

    assert hasattr(m2.fs.tc, "inlet")
    assert hasattr(m2.fs.tc, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.tc.inlet:
        assert k in [0]
    for k in m2.fs.tc.outlet:
        assert k in [(0, "a"), (0, "b")]


def test_post_transform_build_num_outlet2(m2):
    m2.fs.tc.config.num_outlets = 2
    m2.fs.tc.post_transform_build()

    assert hasattr(m2.fs.tc, "inlet")
    assert hasattr(m2.fs.tc, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.tc.inlet:
        assert k in [0]
    for k in m2.fs.tc.outlet:
        assert k in [(0, "1"), (0, "2")]


# -----------------------------------------------------------------------------
# Test model_checks
def test_model_check_heater_true():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.tc = TemperatureChanger(property_package=m.fs.props,
                                 heater=True)
    m.fs.post_transform_build()
    m.fs.tc.model_check()


def test_model_check_heater_false():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.tc = TemperatureChanger(property_package=m.fs.props,
                                 heater=False)
    m.fs.post_transform_build()
    m.fs.tc.model_check()


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.tc = TemperatureChanger(property_package=m.fs.props)
    m.fs.post_transform_build()
    
    m.fs.tc.inlet[:].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
    m.fs.tc.inlet[:].vars["temperature"].fix(866.5)
    m.fs.tc.inlet[:].vars["pressure"].fix(2.97e7)

    m.fs.tc.deltaT.fix(10.0)

    assert degrees_of_freedom(m) == 0

    m.fs.tc.initialize()

    assert (pytest.approx(27500, abs=1e-2) ==
            m.fs.tc.outlet[0].vars["flow_mol_comp"]["H2O"].value)
    assert (pytest.approx(876.5, abs=1e-2) ==
            m.fs.tc.outlet[0].vars["temperature"].value)
    assert (pytest.approx(29700000.0, abs=1e-2) ==
            m.fs.tc.outlet[0].vars["pressure"].value)
    assert (pytest.approx(15862193.476, abs=1e-2) ==
            m.fs.tc.heat[0].value)
